# -*- encoding:utf-8 -*-
"""示例ump单混特征模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureDeg, AbuFeaturePrice, AbuFeatureWave, AbuFeatureAtr
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_main_make_xy, BuyUmpMixin
from .ABuUmpMainBase import AbuUmpMainBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpMainMul(AbuUmpMainBase, BuyUmpMixin):
    """主裁单混特征类，AbuUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpMulFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

            在__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_main_make_xy
        def make_xy(self, **kwarg):
            """
            make_xy通过装饰器ump_main_make_xy进行二次包装
            这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
            :param kwarg: ump_main_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                          kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

            :return: self.order_has_ret中使用filter选取特征列形成mul_df
                     ump_main_make_xy装饰器在make_xy返回mul_df后做转换matrix，形成x，y等工作
            """

            # regex='result|buy_deg_ang21|buy_price_rank252|buy_wave_score3|buy_atr_std'
            regex = 'result|{}|{}|{}|{}'.format(AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                                                AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                                                AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                                                AbuFeatureAtr().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1])
            # noinspection PyUnresolvedReferences
            mul_df = self.order_has_ret.filter(regex=regex)
            """
                eg: mul_df
                            result  buy_deg_ang21  buy_price_rank252  buy_wave_score3  \
                2014-09-24       0          1.818              1.000            0.218
                2014-10-24       0         -1.089              1.000            0.449
                2014-10-29       1          1.980              1.000            0.329
                2014-10-29       1          6.507              0.750           -0.203
                2014-10-29       1          7.046              0.982           -0.173
                2014-10-29       1          1.980              1.000            0.329
                2014-11-03       1         10.125              1.000           -0.172
                2014-11-11       1         12.397              0.808            0.001
                2014-11-12       0          6.671              0.560           -0.509
                2014-11-26       0         12.494              0.762            0.015
                ...            ...            ...                ...              ...
                2016-03-14       1          9.324              0.444            0.948
                2016-03-14       0          5.201              0.623            0.788
                2016-03-30       0          4.070              0.536            0.702
                2016-04-04       1          5.618              0.190            0.743
                2016-04-13       0          4.457              0.270            0.651
                2016-04-14       0          0.733              0.940            0.170
                2016-04-15       0          1.162              0.631            0.800
                2016-04-15       0          1.162              0.631            0.800
                2016-04-29       1         -1.115              1.000            0.918
                2016-04-29       1         -1.115              1.000            0.918

                            buy_atr_std
                2014-09-24        0.226
                2014-10-24        0.146
                2014-10-29        0.538
                2014-10-29        0.558
                2014-10-29        0.063
                2014-10-29        0.538
                2014-11-03       -0.002
                2014-11-11        0.171
                2014-11-12        0.093
                2014-11-26        0.812
                ...                 ...
                2016-03-14        0.126
                2016-03-14        0.663
                2016-03-30       -0.008
                2016-04-04        0.035
                2016-04-13        0.262
                2016-04-14        0.049
                2016-04-15        0.135
                2016-04-15        0.135
                2016-04-29        0.838
                2016-04-29        0.838
            """
            return mul_df

    def get_predict_col(self):
        """
        主裁单混特征keys：['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        :return: ['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        """

        return [AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1],
                AbuFeatureAtr().get_feature_ump_keys(ump_cls=AbuUmpMainMul)[-1]]

    def get_fiter_class(self):
        """
        主裁单混特征返回的AbuMLPd子类：AbuUmpMainMul.UmpMulFiter
        :return: AbuUmpMainMul.UmpMulFiter
        """
        return AbuUmpMainMul.UmpMulFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'mul_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'mul_main'



# -*- encoding:utf-8 -*-
"""
    择时并行多任务调度模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd

from .ABuPickTimeExecute import do_symbols_with_same_factors
from ..CoreBu.ABuEnvProcess import AbuEnvProcess
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataFetchMode
from ..MarketBu.ABuMarket import split_k_market
from ..TradeBu import ABuTradeExecute
from ..TradeBu.ABuKLManager import AbuKLManager
from ..CoreBu.ABuParallel import delayed, Parallel

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuPickTimeMaster(object):
    """择时并行多任务调度类"""

    @classmethod
    def do_symbols_with_same_factors_process(cls, target_symbols, benchmark, buy_factors, sell_factors, capital,
                                             kl_pd_manager=None,
                                             n_process_kl=ABuEnv.g_cpu_cnt * 2 if ABuEnv.g_is_mac_os
                                             else ABuEnv.g_cpu_cnt,
                                             n_process_pick_time=ABuEnv.g_cpu_cnt):
        """
        将多个交易对象拆解为多份交易对象序列，多任务并行完成择时工作
        :param target_symbols: 多个择时交易对象序列
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param buy_factors: 买入因子序列
        :param sell_factors: 卖出因子序列
        :param capital: AbuCapital实例对象
        :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
        :param n_process_kl: 控制金融时间序列管理对象内部启动n_process_kl进程获取金融序列数据
        :param n_process_pick_time: 控制择时操作并行任务数量
        :return:
        """

        if kl_pd_manager is None:
            kl_pd_manager = AbuKLManager(benchmark, capital)
            # 一次性在主进程中执行多进程获取k线数据，全部放入kl_pd_manager中，内部启动n_process_kl个进程执行
            kl_pd_manager.batch_get_pick_time_kl_pd(target_symbols, n_process=n_process_kl)

        if n_process_pick_time <= 0:
            # 因为下面要根据n_process_pick_time来split_k_market
            n_process_pick_time = ABuEnv.g_cpu_cnt

        # 将target_symbols切割为n_process_pick_time个子序列，这样可以每个进程处理一个子序列
        process_symbols = split_k_market(n_process_pick_time, market_symbols=target_symbols)

        # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
        n_process_pick_time = len(process_symbols)

        parallel = Parallel(
            n_jobs=n_process_pick_time, verbose=0, pre_dispatch='2*n_jobs')

        tmp_fetch_mode = ABuEnv.g_data_fetch_mode
        if n_process_pick_time > 1:
            # TODO 需要区分hdf5和csv不同存贮情况，csv存贮模式下可以并行读写
            # 因为上面已经并行或者单进程进行数据采集kl_pd_manager，之后的并行，为确保hdf5不会多进程读写设置LOCAL
            ABuEnv.g_data_fetch_mode == EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL

        # do_symbols_with_same_factors被装饰器add_process_env_sig装饰，需要进程间内存拷贝对象AbuEnvProcess
        p_nev = AbuEnvProcess()
        # 每个并行的进程通过do_symbols_with_same_factors及自己独立的子序列独立工作，注意kl_pd_manager装载了所有需要的数据
        out = parallel(delayed(do_symbols_with_same_factors)(choice_symbols, benchmark, buy_factors, sell_factors,
                                                             capital, apply_capital=False,
                                                             kl_pd_manager=kl_pd_manager, env=p_nev)
                       for choice_symbols in process_symbols)
        # 择时并行结束后恢复之前的数据获取模式
        ABuEnv.g_data_fetch_mode = tmp_fetch_mode
        orders_pd = None
        action_pd = None
        all_fit_symbols_cnt = 0
        for sub_out in out:
            # 将每个子序列进程的处理结果进行合并
            sub_orders_pd, sub_action_pd, sub_all_fit_symbols_cnt = sub_out
            orders_pd = sub_orders_pd if orders_pd is None else pd.concat([orders_pd, sub_orders_pd])
            action_pd = sub_action_pd if action_pd is None else pd.concat([action_pd, sub_action_pd])
            all_fit_symbols_cnt += sub_all_fit_symbols_cnt

        if orders_pd is not None and action_pd is not None:
            # 将合并后的结果按照时间及行为进行排序
            # noinspection PyUnresolvedReferences
            action_pd = action_pd.sort_values(['Date', 'action'])
            action_pd.index = np.arange(0, action_pd.shape[0])
            # noinspection PyUnresolvedReferences
            orders_pd = orders_pd.sort_values(['buy_date'])
            # 最后将所有的action作用在资金上，生成资金时序，及判断是否能买入
            ABuTradeExecute.apply_action_to_capital(capital, action_pd, kl_pd_manager)

        return orders_pd, action_pd, all_fit_symbols_cnt



# -*- encoding:utf-8 -*-
"""
    示例仓位管理：kelly仓位管理模块
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from .ABuPositionBase import AbuPositionBase
from . import ABuPositionBase


class AbuKellyPosition(AbuPositionBase):
    """示例kelly仓位管理类"""

    def fit_position(self, factor_object):
        """
        fit_position计算的结果是买入多少个单位（股，手，顿，合约）
        需要factor_object策略因子对象通过历史回测统计胜率，期望收益，期望亏损，
        并设置构造当前factor_object对象，通过kelly公司计算仓位
        :param factor_object: ABuFactorBuyBases子类实例对象
        :return:买入多少个单位（股，手，顿，合约）
        """

        # 检测择时策略因子对象有没有设置设置胜率，期望收益，期望亏损，详情查阅ABuFactorBuyBase
        if not hasattr(factor_object, 'win_rate'):
            raise RuntimeError('AbuKellyPosition need factor_object has  win_rate')
        if not hasattr(factor_object, 'gains_mean'):
            raise RuntimeError('AbuKellyPosition need factor_object has  gains_mean')
        if not hasattr(factor_object, 'losses_mean'):
            raise RuntimeError('AbuKellyPosition need factor_object has  losses_mean')

        # 胜率
        win_rate = factor_object.win_rate
        # 败率
        loss_rate = 1 - win_rate
        # 平均获利期望
        gains_mean = factor_object.gains_mean
        # 平均亏损期望
        losses_mean = factor_object.losses_mean
        # kelly计算出仓位比例
        kelly_pos = win_rate - loss_rate / (gains_mean / losses_mean)
        # 最大仓位限制
        kelly_pos = ABuPositionBase.g_pos_max if kelly_pos > ABuPositionBase.g_pos_max else kelly_pos
        # 结果是买入多少个单位（股，手，顿，合约）
        return self.read_cash * kelly_pos / self.bp * self.deposit_rate



# -*- encoding:utf-8 -*-
"""
    检查范围的函数
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np
from abc import ABCMeta

from ..CoreBu.ABuFixes import six

__author__ = '夜猫'
__weixin__ = 'abu_quant'


class CheckError(six.with_metaclass(ABCMeta, TypeError)):
    """Check失败的Error类型"""
    pass


def bound_check(bound):
    """
    制作检查数值范围的check_fail函数
    """
    (min_val, max_val) = bound
    # 转换None到inf
    min_val = -np.inf if min_val is None else min_val
    max_val = np.inf if max_val is None else max_val
    # 准备错误message
    error_msg = "function expected a return value inclusively between %s and %s" % (min_val, max_val)

    def _check(value):
        """范围检查逻辑"""
        return not (min_val <= value <= max_val)

    def _bound_check(val):
        """
        检查数值范围的函数；检查失败raise CheckError
        """
        if _check(val):
            raise CheckError(error_msg)
        else:
            return val

    return _bound_check


def bound_valid_and_check(bound):
    """
    检查bound的输入参数格式；失败raise TypeError
    传入参数形式应为`` (min_value, max_value)``.
    """

    def valid_bound(t):
        # 检查bound传入参数格式
        return (
            isinstance(t, tuple)
            and len(t) == 2
            and t != (None, None)
        )

    if not valid_bound(bound):
        # 参数格式错误
        raise TypeError(
            "function expected a tuple of bounds,"
            "but got {} instead.".format(bound)
        )
    return bound_check(bound)


def subset_check(subset):
    """
    制作检查是否子集的check函数
    """

    def _check(arg_val):
        """
        检查数值是否子集的check函数；检查失败raise CheckError
        """
        if subset is not None and arg_val not in set(subset):
            raise CheckError(
                'Value {} is not the subset of {}'.format(arg_val, str(subset))
            )
        return arg_val

    return _check


def type_check(arg_ty):
    """
    制作检查数值类型的check函数
    """

    def _check(arg_val):
        """
        检查数值类型的check函数；检查失败raise CheckError
        """
        if arg_ty and not isinstance(arg_val, arg_ty):
            raise CheckError(
                'Value {} is not {}'.format(arg_val, arg_ty)
            )
        return arg_val

    return _check



# -*- encoding:utf-8 -*-
"""
    金融时间序列管理模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging

from ..TradeBu import AbuBenchmark
from ..UtilBu import ABuDateUtil
from ..CoreBu.ABuEnv import EMarketDataSplitMode, EMarketDataFetchMode
from ..MarketBu import ABuSymbolPd
from ..MarketBu.ABuMarket import split_k_market
from ..CoreBu.ABuEnvProcess import add_process_env_sig, AbuEnvProcess
from ..CoreBu.ABuParallel import delayed, Parallel
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EDataCacheType
from ..UtilBu.ABuProgress import AbuMulPidProgress
from ..UtilBu.ABuFileUtil import batch_h5s
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnusedLocal
@add_process_env_sig
def gen_dict_pick_time_kl_pd(target_symbols, capital, benchmark):
    """
    在AbuKLManager中batch_get_pick_time_kl_pd批量获取择时时间序列中使用做为并行多进程委托方法
    :param target_symbols: 请求的symbol
    :param capital: 资金类AbuCapital实例化对象 （实现中暂时不使用其中信息）
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    """

    # 构建的返回时间序列交易数据组成的字典
    pick_kl_pd_dict = dict()

    # 为batch_h5s装饰器准备参数，详见batch_h5装饰器实现
    h5s_fn = None
    if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_HDF5 and ABuEnv.g_data_fetch_mode == \
            EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
        # 存储使用hdf5且使用本地数据模式才赋予h5s_fn路径
        # noinspection PyProtectedMember
        h5s_fn = ABuEnv.g_project_kl_df_data_example if ABuEnv._g_enable_example_env_ipython \
            else ABuEnv.g_project_kl_df_data

    @batch_h5s(h5s_fn)
    def _batch_gen_dict_pick_time_kl_pd():
        # 启动多进程进度条
        with AbuMulPidProgress(len(target_symbols), 'gen kl_pd complete') as progress:
            for epoch, target_symbol in enumerate(target_symbols):
                progress.show(epoch + 1)
                # 迭代target_symbols，获取对应时间交易序列
                kl_pd = ABuSymbolPd.make_kl_df(target_symbol, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                               benchmark=benchmark, n_folds=benchmark.n_folds)
                # 以target_symbol为key将时间金融序列kl_pd添加到返回字典中
                pick_kl_pd_dict[target_symbol] = kl_pd
    _batch_gen_dict_pick_time_kl_pd()
    return pick_kl_pd_dict


class AbuKLManager(object):
    """金融时间序列管理类"""

    def __init__(self, benchmark, capital):
        """
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param capital: 资金类AbuCapital实例化对象
        """
        self.benchmark = benchmark
        self.capital = capital
        # 选股时间交易序列字典
        pick_stock_kl_pd_dict = dict()
        # 择时时间交易序列字典
        pick_time_kl_pd_dict = dict()
        # 类字典pick_kl_pd_dict将选股和择时字典包起来
        self.pick_kl_pd_dict = {'pick_stock': pick_stock_kl_pd_dict, 'pick_time': pick_time_kl_pd_dict}

    def __str__(self):
        """打印对象显示：pick_stock + pick_time keys, 即所有symbol信息"""
        keys = set(self.pick_kl_pd_dict['pick_stock'].keys()) | set(self.pick_kl_pd_dict['pick_time'].keys())
        return 'pick_stock + pick_time keys :{}'.format(keys)

    __repr__ = __str__

    def __len__(self):
        """对象长度：选股字典长度 + 择时字典长度"""
        return len(self.pick_kl_pd_dict['pick_stock']) + len(self.pick_kl_pd_dict['pick_time'])

    def __contains__(self, item):
        """成员测试：在择时字典中或者在选股字典中"""
        return item in self.pick_kl_pd_dict['pick_stock'] or item in self.pick_kl_pd_dict['pick_time']

    def __missing__(self, key):
        """对象缺失：需要根据key使用code_to_symbol进行fetch数据，暂未实现"""
        # TODO 需要根据key使用code_to_symbol进行fetch数据
        raise NotImplementedError('TODO AbuKLManager __missing__')

    def __getitem__(self, key):
        """索引获取：尝试分别从选股字典，择时字典中查询，返回两个字典的查询结果"""
        pick_stock_item = None
        if key in self.pick_kl_pd_dict['pick_stock']:
            pick_stock_item = self.pick_kl_pd_dict['pick_stock'][key]
        pick_time_item = None
        if key in self.pick_kl_pd_dict['pick_time']:
            pick_time_item = self.pick_kl_pd_dict['pick_time'][key]
        return pick_stock_item, pick_time_item

    def __setitem__(self, key, value):
        """索引设置：抛错误，即不准许外部设置"""
        raise AttributeError("AbuKLManager set value!!!")

    def _fetch_pick_stock_kl_pd(self, xd, target_symbol):
        """
        根据选股周期和symbol获取选股时段金融时间序列，相对择时金融时间序列获取要复杂，
        因为要根据条件构造选股时段benchmark，且在类变量中存储选股时段benchmark
        :param xd: 选股周期（默认一年的交易日长度）
        :param target_symbol: 选股symbol
        :return: 选股时段金融时间序列
        """

        # 从设置的择时benchmark中取第一个日期即为选股时段最后一个日期
        end = ABuDateUtil.timestamp_to_str(self.benchmark.kl_pd.index[0])

        if xd == ABuEnv.g_market_trade_year:
            # 一般都是默认的1年，不需要使用begin_date提高效率
            n_folds = 1
            pre_bc_key = 'pre_benchmark_{}'.format(n_folds)
            start = None
        else:
            # 1年除1年交易日数量，浮点数n_folds eg: 0.88
            n_folds = float(xd / ABuEnv.g_market_trade_year)
            # 为了计算start，xd的单位是交易日，换算为自然日
            delay_day = 365 * n_folds
            start = ABuDateUtil.begin_date(delay_day, date_str=end, fix=False)
            # 根据选股start，end拼接选股类变量key，eg：pre_benchmark_2011-09-09_2016-07-26
            pre_bc_key = 'pre_benchmark_{}-{}'.format(start, end)
        if hasattr(self, pre_bc_key):
            # 从类变量中直接获取选股benchmark，eg: self.pre_benchmark_2011-09-09_2016-07-26
            pre_benchmark = getattr(self, pre_bc_key)
        else:
            # 类变量中没有，实例一个AbuBenchmark，根据n_folds和end获取benchmark选股时段
            pre_benchmark = AbuBenchmark(n_folds=n_folds, start=start, end=end)
            # 类变量设置选股时段benchmark
            setattr(self, pre_bc_key, pre_benchmark)
        # 以选股时段benchmark做为参数，获取选股时段对应symbol的金融时间序列
        return ABuSymbolPd.make_kl_df(target_symbol, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                      benchmark=pre_benchmark, n_folds=pre_benchmark.n_folds, start=start, end=end)

    def _fetch_pick_time_kl_pd(self, target_symbol):
        """获取择时时段金融时间序列"""
        return ABuSymbolPd.make_kl_df(target_symbol, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                      benchmark=self.benchmark, n_folds=self.benchmark.n_folds)

    def get_pick_time_kl_pd(self, target_symbol):
        """对外获取择时时段金融时间序列，首先在内部择时字典中寻找，没找到使用_fetch_pick_time_kl_pd获取，且保存择时字典"""
        if target_symbol in self.pick_kl_pd_dict['pick_time']:
            kl_pd = self.pick_kl_pd_dict['pick_time'][target_symbol]
            if kl_pd is not None:
                # 因为在多进程的时候拷贝会丢失name信息
                kl_pd.name = target_symbol
            return kl_pd
        # 字典中每找到，进行fetch，获取后保存在择时字典中
        kl_pd = self._fetch_pick_time_kl_pd(target_symbol)
        self.pick_kl_pd_dict['pick_time'][target_symbol] = kl_pd
        return kl_pd

    def filter_pick_time_choice_symbols(self, choice_symbols):
        """
        使用filter筛选出choice_symbols中的symbol对应的择时时间序列不在内部择时字典中的symbol序列
        :param choice_symbols: 支持迭代的symbol序列
        :return: 不在内部择时字典中的symbol序列
        """
        return list(filter(lambda target_symbol: target_symbol not in self.pick_kl_pd_dict['pick_time'],
                           choice_symbols))

    def batch_get_pick_time_kl_pd(self, choice_symbols, n_process=ABuEnv.g_cpu_cnt):
        """
        统一批量获取择时金融时间序列获保存在内部的择时字典中，以多进程并行方式运行
        :param choice_symbols: 支持迭代的symbol序列
        :param n_process: 择时金融时间序列获取并行启动的进程数，默认16个，属于io操作多，所以没有考虑cpu数量
        :return:
        """
        if len(choice_symbols) == 0:
            return

        if n_process <= 0:
            # 因为下面要n_process > 1做判断而且要根据n_process来split_k_market
            n_process = ABuEnv.g_cpu_cnt

        # TODO 需要区分hdf5和csv不同存贮情况，csv存贮模式下可以并行读写
        # 只有E_DATA_FETCH_FORCE_LOCAL才进行多任务模式，否则回滚到单进程模式n_process = 1
        if n_process > 1 and ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
            # 1. hdf5多进程还容易写坏数据
            # 2. MAC OS 10.9 之后并行联网＋numpy 系统bug crash，卡死等问题
            logging.info('batch get only support E_DATA_FETCH_FORCE_LOCAL for Parallel!')
            n_process = 1

        # 根据输入的choice_symbols和要并行的进程数，分配symbol到n_process个进程中
        process_symbols = split_k_market(n_process, market_symbols=choice_symbols)

        # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
        if n_process > 1:
            n_process = len(process_symbols)

        parallel = Parallel(
            n_jobs=n_process, verbose=0, pre_dispatch='2*n_jobs')

        # gen_dict_pick_time_kl_pd被装饰器add_process_env_sig装饰，需要进程间内存拷贝对象AbuEnvProcess，详ABuEnvProcess.py
        p_nev = AbuEnvProcess()
        # 开始并行任务执行
        out_pick_kl_pd_dict = parallel(delayed(gen_dict_pick_time_kl_pd)(target_symbols, self.capital, self.benchmark,
                                                                         env=p_nev)
                                       for target_symbols in process_symbols)

        for pick_kl_pd_dict in out_pick_kl_pd_dict:
            # 迭代多任务组成的out_pick_kl_pd_dict，分别更新保存在内部的择时字典中
            self.pick_kl_pd_dict['pick_time'].update(pick_kl_pd_dict)

    def get_pick_stock_kl_pd(self, target_symbol, xd=ABuEnv.g_market_trade_year,
                             min_xd=int(ABuEnv.g_market_trade_year / 2)):
        """
        对外获取选股时段金融时间序列，首先在内部择时字典中寻找，没找到使用_fetch_pick_stock_kl_pd获取，且保存选股字典
        :param target_symbol: 选股symbol
        :param xd: 选股周期（默认一年的交易日长度）
        :param min_xd: 对fetch的选股金融序列进行过滤参数，即最小金融序列长度
        :return:
        """

        if target_symbol in self.pick_kl_pd_dict['pick_stock']:
            xd_dict = self.pick_kl_pd_dict['pick_stock'][target_symbol]
            if xd in xd_dict:
                # 缓存中找到形如：self.pick_kl_pd_dict['pick_stock']['usTSLA']['252']
                # noinspection PyTypeChecker
                kl_pd = xd_dict[xd]
                if kl_pd is not None:
                    # 因为在多进程的时候深拷贝会丢失name
                    kl_pd.name = target_symbol
                return kl_pd

        # 字典中每找到，进行fetch
        kl_pd = self._fetch_pick_stock_kl_pd(xd, target_symbol)
        """选股字典是三层字典结构，比择时字典多一层，因为有选股周期做为第三层字典的key"""
        if kl_pd is None or kl_pd.shape[0] == 0:
            self.pick_kl_pd_dict['pick_stock'][target_symbol] = {xd: None}
            return None

        """由于_fetch_pick_stock_kl_pd中获取kl_pd使用了标尺模式，所以这里的min_xd要设置大于标尺才有实际意义"""
        if kl_pd.shape[0] < min_xd:
            # 如果时间序列有数据但是 < min_xd, 抛弃数据直接{xd: None}
            self.pick_kl_pd_dict['pick_stock'][target_symbol] = {xd: None}
            return None
        # 第三层字典{xd: kl_pd}
        self.pick_kl_pd_dict['pick_stock'][target_symbol] = {xd: kl_pd}
        return kl_pd



# -*- encoding:utf-8 -*-
"""针对交易回测结果存储，读取模块"""

import os
from collections import namedtuple
from enum import Enum

from ..CoreBu import ABuEnv
from ..UtilBu import ABuFileUtil


# noinspection PyClassHasNoInit
class AbuResultTuple(namedtuple('AbuResultTuple',
                                ('orders_pd',
                                 'action_pd',
                                 'capital',
                                 'benchmark'))):
    """
        使用abu.run_loop_back返回的nametuple对象：

        orders_pd：回测结果生成的交易订单构成的pd.DataFrame对象
        action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        capital:   资金类AbuCapital实例化对象
        benchmark: 交易基准对象，AbuBenchmark实例对象
    """
    __slots__ = ()

    def __repr__(self):
        """打印对象显示：orders_pd.info, action_pd.info, capital, benchmark"""
        return "orders_pd:{}\naction_pd:{}\ncapital:{}\nbenchmark:{}".format(
            self.orders_pd.info(),
            self.action_pd.info(),
            self.capital, self.benchmark)


class EStoreAbu(Enum):
    """保存回测结果的enum类型"""

    """保存普通类型，存储文件后缀为空"""
    E_STORE_NORMAL = 0

    """保存训练集回测，存储文件后缀为train"""
    E_STORE_TRAIN = 1
    """保存测试集交易回测，存储文件后缀为test"""
    E_STORE_TEST = 2

    """保存测试集交易使用主裁ump进行回测，存储文件后缀为test_ump"""
    E_STORE_TEST_UMP = 3
    """保存测试集交易使用主裁＋边裁ump进行回测，存储文件后缀为test_ump_with_edge"""
    E_STORE_TEST_UMP_WITH_EDGE = 4

    """保存测回测，存储文件后缀为自定义字符串"""
    E_STORE_CUSTOM_NAME = 5


def store_abu_result_tuple(abu_result_tuple, n_folds, store_type=None, custom_name=None):
    """
    保存abu.run_loop_back的回测结果AbuResultTuple对象，根据n_folds，store_type参数
    来定义存储的文件名称

    :param abu_result_tuple: AbuResultTuple对象类型
    :param n_folds: 回测执行了几年，只影响存贮文件名
    :param store_type: 回测保存类型EStoreAbu类型，只影响存贮文件名
    :param custom_name: 如果store_type=EStoreAbu.E_STORE_CUSTOM_NAME时需要的自定义文件名称
    """
    fn_root = ABuEnv.g_project_cache_dir
    fn_head = 'n{}_'.format(n_folds)

    # 根据EStoreAbu来决定fn_head
    if store_type == EStoreAbu.E_STORE_TEST:
        fn_head += 'test'
    elif store_type == EStoreAbu.E_STORE_TEST_UMP:
        fn_head += 'test_ump'
    elif store_type == EStoreAbu.E_STORE_TEST_UMP_WITH_EDGE:
        fn_head += 'test_ump_with_edge'
    elif store_type == EStoreAbu.E_STORE_TRAIN:
        fn_head += 'train'
    elif store_type == EStoreAbu.E_STORE_CUSTOM_NAME:
        fn_head += custom_name

    # eg: n2_test_orders_pd
    key = fn_head + '_orders_pd'
    fn = os.path.join(fn_root, key)
    ABuFileUtil.ensure_dir(fn)
    # abu_result_tuple.orders_pd使用dump_hdf5存储
    ABuFileUtil.dump_hdf5(fn, abu_result_tuple.orders_pd, key)

    # eg: n2_test_action_pd
    key = fn_head + '_action_pd'
    fn = os.path.join(fn_root, key)
    # abu_result_tuple.action_pd使用dump_hdf5存储
    ABuFileUtil.dump_hdf5(fn, abu_result_tuple.action_pd, key)

    # eg: n2_test_capital
    fn = os.path.join(fn_root, fn_head + '_capital')
    # abu_result_tuple.capital使用dump_pickle存储AbuCapital对象
    ABuFileUtil.dump_pickle(abu_result_tuple.capital, fn)

    # eg: n2_test_benchmark
    fn = os.path.join(fn_root, fn_head + '_benchmark')
    # abu_result_tuple.benchmark使用dump_pickle存储AbuBenchmark对象
    ABuFileUtil.dump_pickle(abu_result_tuple.benchmark, fn)


def load_abu_result_tuple(n_folds, store_type, custom_name=None):
    """
    读取使用store_abu_result_tuple保存的回测结果，根据n_folds，store_type参数
    来定义读取的文件名称，依次读取orders_pd，action_pd，capital，benchmark后构造
    AbuResultTuple对象返回

    :param n_folds: 回测执行了几年，只影响读取的文件名
    :param store_type: 回测保存类型EStoreAbu类型，只影响读取的文件名
    :param custom_name: 如果store_type=EStoreAbu.E_STORE_CUSTOM_NAME时需要的自定义文件名称
    :return: AbuResultTuple对象
    """
    fn_root = ABuEnv.g_project_cache_dir
    fn_head = 'n{}_'.format(n_folds)
    # 根据EStoreAbu来决定fn_head
    if store_type == EStoreAbu.E_STORE_TRAIN:
        fn_head += 'train'
    elif store_type == EStoreAbu.E_STORE_TEST:
        fn_head += 'test'
    elif store_type == EStoreAbu.E_STORE_TEST_UMP:
        fn_head += 'test_ump'
    elif store_type == EStoreAbu.E_STORE_TEST_UMP_WITH_EDGE:
        fn_head += 'test_ump_with_edge'
    elif store_type == EStoreAbu.E_STORE_CUSTOM_NAME:
        fn_head += custom_name
    elif store_type != EStoreAbu.E_STORE_NORMAL:
        raise ValueError('store_type error!!!')

    # eg: n2_test_orders_pd
    key = fn_head + '_orders_pd'
    fn = os.path.join(fn_root, key)
    # load_hdf5读取pd.DataFrame对象orders_pd
    orders_pd = ABuFileUtil.load_hdf5(fn, key)

    # eg: n2_test_action_pd
    key = fn_head + '_action_pd'
    fn = os.path.join(fn_root, key)
    # load_hdf5读取pd.DataFrame对象action_pd
    action_pd = ABuFileUtil.load_hdf5(fn, key)

    fn = os.path.join(fn_root, fn_head + '_capital')
    # load_pickle读取AbuCapital对象capital，AbuCapital混入了类PickleStateMixin
    capital = ABuFileUtil.load_pickle(fn)

    fn = os.path.join(fn_root, fn_head + '_benchmark')
    # load_pickle读取AbuBenchmark对象benchmark，AbuBenchmark混入了类PickleStateMixin
    benchmark = ABuFileUtil.load_pickle(fn)
    # 构建返回AbuResultTuple对象
    return AbuResultTuple(orders_pd, action_pd, capital, benchmark)



# -*- encoding:utf-8 -*-
"""
    拟合工具模块
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import math
import logging

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from statsmodels import api as sm, regression
from sklearn import metrics

from ..CoreBu import ABuEnv
from ..CoreBu.ABuPdHelper import pd_rolling_mean
from ..UtilBu.ABuStatsUtil import euclidean_distance_xy, manhattan_distances_xy, cosine_distances_xy


log_func = logging.info if ABuEnv.g_is_ipython else print


def regress_xy(x, y, mode=True, zoom=False, show=False):
    """
    使用statsmodels.regression.linear_model进行简单拟合操作，返回model和y_fit
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param mode: 是否需要mode结果，在只需要y_fit且效率需要高时应设置False, 效率差异：
                 mode=False: 1000 loops, best of 3: 778 µs per loop
                 mode=True:  1000 loops, best of 3: 1.23 ms per loop
    :param zoom: 是否缩放x,y
    :param show: 是否可视化结果
    :return: model, y_fit, 如果mode=False，返回的model=None
    """
    if zoom:
        # 将y值 zoom到与x一个级别，不可用ABuScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        # TODO ABuScalerUtil中添加使用固定轴进行缩放的功能
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y

    if mode:
        # 加常数1列
        x = sm.add_constant(x)
        model = regression.linear_model.OLS(y, x).fit()

        intercept = model.params[0]
        rad = model.params[1]
        # y = kx + b, x取x[:, 1]，因为add_constant
        y_fit = x[:, 1] * rad + intercept
    else:
        # noinspection PyCallingNonCallable
        y_fit = np.polynomial.Chebyshev.fit(x, y, 1)(x)
        model = None
    if show:
        # 取-1因为有OLS add_constant和Chebyshev没有add_constant的两种情况
        x_plot = x[:, -1]
        # 绘制x， y
        plt.plot(x_plot, y)
        # 绘制x， 拟合的y
        plt.plot(x_plot, y_fit)
        plt.show()

        # 再使用sns绘制，对比拟合结果
        sns.regplot(x=x_plot, y=y)
        plt.show()
    return model, y_fit


def regress_y(y, mode=True, zoom=False, show=False):
    """
    使用statsmodels.regression.linear_model进行简单拟合操作, 参数中只提供y序列，
    x使用np.arange(0, len(y))填充
    :param y: 可迭代序列
    :param mode: 是否需要mode结果，在只需要y_fit且效率需要高时应设置False, 效率差异：
             mode=False: 1000 loops, best of 3: 778 µs per loop
             mode=True:  1000 loops, best of 3: 1.23 ms per loop
    :param zoom: 是否缩放x,y
    :param show: 是否可视化结果
    :return: model, y_fit, 如果mode=False，返回的model=None
    """
    x = np.arange(0, len(y))
    return regress_xy(x, y, mode=mode, zoom=zoom, show=show)


def calc_regress_deg(y, show=True):
    """
    将y值 zoom到与x一个级别，之后再fit出弧度转成角度
    1 多个股票的趋势比较提供量化基础，只要同一个时间范围，就可以比较
    2 接近视觉感受到的角度
    :param y:  可迭代序列
    :param show: 是否可视化结果
    :return: deg角度float值
    """
    # 将y值 zoom到与x一个级别
    model, _ = regress_y(y, mode=True, zoom=True, show=show)
    rad = model.params[1]
    # fit出弧度转成角度
    deg = np.rad2deg(rad)
    return deg


def regress_xy_polynomial(x, y, poly=1, zoom=False, show=False):
    """
    多项式拟合, 根据参数poly决定，返回拟合后的y_fit
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param poly: 几次拟合参数，int
    :param zoom: 是否对数据进行缩放
    :param show: 是否可视化显示拟合结果
    :return: y_fit
    """
    if zoom:
        # 将y值 zoom到与x一个级别，不可用ABuScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y

    polynomial = np.polynomial.Chebyshev.fit(x, y, poly)
    # noinspection PyCallingNonCallable
    y_fit = polynomial(x)

    if show:
        # 可视化显示拟合结果
        plt.plot(x, y)
        plt.plot(x, y_fit)
        plt.title('{} poly zoom ={}'.format(poly, zoom))
        plt.show()

    return y_fit


def regress_y_polynomial(y, poly=1, zoom=False, show=False):
    """
    套接regress_xy_polynomial操作, 参数中只提供y序列，x使用np.arange(0, len(y))填充
    :param y: 可迭代序列
    :param poly: 几次拟合参数，int
    :param zoom: 是否对数据进行缩放
    :param show: 是否可视化显示拟合结果
    :return: y_fit
    """
    x = np.arange(0, len(y))
    return regress_xy_polynomial(x, y, poly=poly, zoom=zoom, show=show)


def metrics_mae(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的MAE：
         MAE = sum(np.abs(y - y_fit)) / len(y)
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出mae值
    :return: 返回mae值，float
    """
    mae = metrics.mean_absolute_error(y, y_fit)
    if show:
        log_func('MAE={}'.format(mae))
    return mae


def metrics_mse(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的MSE：
         MSE = sum(np.square(y - y_fit)) / len(y)
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出MSE值
    :return: 返回MSE值，float
    """
    mse = metrics.mean_squared_error(y, y_fit)
    if show:
        log_func('MSE={}'.format(mse))
    return mse


def metrics_rmse(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的RMSE：
         RMSE = np.sqrt(sum(np.square(y - y_fit)) / len(y))
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出RMSE值
    :return: 返回RMSE值，float
    """
    rmse = np.sqrt(metrics.mean_squared_error(y, y_fit))
    if show:
        log_func('RMSE={}'.format(rmse))
    return rmse


def metrics_euclidean(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的euclidean欧式距离(L2范数)
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出欧式距离(L2范数)值
    :return: 返回欧式距离(L2范数)值，float
    """
    euclidean = euclidean_distance_xy(y, y_fit, to_similar=False)
    if show:
        log_func('euclidean={}'.format(euclidean))
    return euclidean


def metrics_manhattan(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的manhattan曼哈顿距离(L1范数)
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出曼哈顿距离(L1范数)值
    :return: 返回曼哈顿距离(L1范数)值，float
    """
    manhattan = manhattan_distances_xy(y, y_fit, to_similar=False)
    if show:
        log_func('manhattan={}'.format(manhattan))
    return manhattan


def metrics_cosine(y, y_fit, show=True):
    """
    度量原始序列和拟合后的y_fit的cosine余弦距离
    :param y: 原始可迭代序列
    :param y_fit: 拟合可迭代序列
    :param show: 是否输出余弦距离值
    :return: 返回曼余弦距离值，float
    """
    cosine = cosine_distances_xy(y, y_fit, to_similar=False)
    if show:
        log_func('cosine={}'.format(cosine))
    return cosine


def valid_poly(y, poly=1, zoom=False, show=True, metrics_func=metrics_rmse):
    """
    验证poly（默认＝1）次多项式拟合回归的趋势曲线是否能代表原始曲线y的走势，
    基础思路：
             1. 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
             eg：
                原始y序列＝504 －> rolling_window = math.ceil(len(y) / 4) = 126
             2. 通过pd_rolling_mean计算出均线的值y_roll_mean
             3  使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
             3. 通过计算regress_xy_polynomial计算多项式拟合回归的趋势曲线y_fit
             4. 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
             5. 如果distance_fit <= distance_mean即代表拟合曲线可以代表原始曲线y的走势
    :param y: 原始可迭代序列
    :param poly: 几次拟合参数，int
    :param zoom: 是否对y数据进行缩放
    :param show: 是否原始曲线y，均线，以及拟合曲线可视化
    :param metrics_func: 度量始y值和均线y_roll_mean的距离和原始y值和
                         拟合回归的趋势曲线y_fit的距离的方法，默认使用metrics_rmse
    :return: 是否poly次拟合曲线可以代表原始曲线y的走势
    """
    valid = False
    x = np.arange(0, len(y))
    if zoom:
        # 将y值 zoom到与x一个级别，不可用ABuScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y
    # 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
    rolling_window = int(math.ceil(len(y) / 4))
    # 通过pd_rolling_mean计算出均线的值y_roll_mean
    y_roll_mean = pd_rolling_mean(y, window=rolling_window, min_periods=1)
    # 使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
    distance_mean = metrics_func(y, y_roll_mean, show=False)

    # 通过计算regress_xy_polynomial计算多项式拟合回归的趋势曲线y_fit, 外面做zoom了所以zoom=False
    y_fit = regress_xy_polynomial(x, y, poly=poly, zoom=False, show=False)
    # 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
    distance_fit = metrics_func(y, y_fit, show=False)
    # 如果distance_fit <= distance_mean即代表拟合曲线可以代表原始曲线y的走势
    if distance_fit <= distance_mean:
        valid = True
    if show:
        # 原始曲线y，均线，以及拟合曲线可视化
        plt.plot(x, y)
        plt.plot(x, y_roll_mean)
        plt.plot(x, y_fit)
        plt.legend(['close', 'rolling window={}'.format(rolling_window), 'y_fit poly={}'.format(poly)])
        plt.show()
        log_func('metrics_func rolling_mean={}, metrics_func y_fit={}'.format(distance_mean, distance_fit))
    return valid


def least_valid_poly(y, zoom=False, show=True, metrics_func=metrics_rmse):
    """
    套接valid_poly，检测至少poly次拟合曲线可以代表原始曲线y的走势
    :param y: 原始可迭代序列
    :param zoom: 是否对y数据进行缩放
    :param show: 是否原始曲线y，均线，以及拟合曲线可视化
    :param metrics_func: 度量始y值和均线y_roll_mean的距离和原始y值和
                         拟合回归的趋势曲线y_fit的距离的方法，默认使用metrics_rmse
    :return: 至少poly次拟合曲线可以代表原始曲线y的走势，int
    """
    poly = 1
    while poly < 100:
        valid = valid_poly(y, poly=poly, zoom=zoom, show=False, metrics_func=metrics_func)
        if valid:
            if show:
                # 这里如果show，就再来了一遍，没在乎效率，在考虑效率情况下不要使用show
                valid_poly(y, poly=poly, zoom=zoom, show=True, metrics_func=metrics_func)
            break
        poly += 1
    return poly


def search_best_poly(y, poly_min=1, poly_max=100, zoom=False, show=True, metrics_func=metrics_rmse):
    """
    寻找poly（1－100）次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势，
    基础思路：
             1. 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
             eg：
                原始y序列＝504 －> rolling_window = math.ceil(len(y) / 4) = 126
             2. 通过pd_rolling_mean计算出均线的值y_roll_mean
             3  使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
             3. 迭代计算1-100poly次regress_xy_polynomial的拟合曲线y_fit
             4. 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
             5. 如果distance_fit <= distance_mean* 0.6即代表拟合曲线可以比较完美的代表原始曲线y的走势，停止迭代
             6. 返回停止迭代时的poly次数
    :param y: 原始可迭代序列
    :param poly_min: 寻找最佳拟合次数的最少次数，eg，2
    :param poly_max: 寻找最佳拟合次数的最多次数，eg：20
    :param zoom: 是否对y数据进行缩放
    :param show: 是否原始曲线y，均线，以及拟合曲线可视化
    :param metrics_func: 度量始y值和均线y_roll_mean的距离和原始y值和
                         拟合回归的趋势曲线y_fit的距离的方法，默认使用metrics_rmse
    :return: 返回停止迭代时的poly次数
    """
    x = np.arange(0, len(y))
    if zoom:
        # 将y值 zoom到与x一个级别，不可用ABuScalerUtil.scaler_xy, 因为不管x > y还y > x都拿 x.max() / y.max()
        zoom_factor = x.max() / y.max()
        y = zoom_factor * y
    # 对原始曲线y进行窗口均线计算，窗口的大小＝ math.ceil(len(y) / 4)
    rolling_window = int(math.ceil(len(y) / 4))
    # 通过pd_rolling_mean计算出均线的值y_roll_mean
    y_roll_mean = pd_rolling_mean(y, window=rolling_window, min_periods=1)
    # 使用metrics_func方法度量原始y值和均线y_roll_mean的距离distance_mean
    distance_mean = metrics_func(y, y_roll_mean, show=False)
    poly = poly_min
    while poly < poly_max:
        # 迭代计算1-100poly次regress_xy_polynomial的拟合曲线y_fit, 外面做zoom了所以zoom=False
        y_fit = regress_xy_polynomial(x, y, poly=poly, zoom=False, show=False)
        # 使用metrics_func方法度量原始y值和拟合回归的趋势曲线y_fit的距离distance_fit
        distance_fit = metrics_func(y, y_fit, show=False)
        if distance_fit <= distance_mean * 0.6:
            # 如果distance_fit <= distance_mean* 0.6即代表拟合曲线可以比较完美的代表原始曲线y的走势，停止迭代
            if show:
                # 原始曲线y，均线，以及拟合曲线可视化
                plt.plot(x, y)
                plt.plot(x, y_roll_mean)
                plt.plot(x, y_fit)
                plt.legend(['close', 'rolling window={}'.format(rolling_window), 'y_fit poly={}'.format(poly)])
                log_func('metrics_func rolling_mean={}, metrics_func y_fit={}'.format(distance_mean, distance_fit))
                plt.show()
            break
        poly += 1
    return poly



# -*- encoding:utf-8 -*-
"""
    检查类，检查函数对象、函数参数、函数返回值
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from toolz import valmap
from functools import wraps

from ..CheckBu.ABuFuncUtil import *
from ..CheckBu.ABuChecks import *
from ..CheckBu.ABuProcessor import arg_process, return_process

__author__ = '夜猫'
__weixin__ = 'abu_quant'


class _NoInstances(six.with_metaclass(ABCMeta, type)):
    """阻止实例化"""

    def __call__(cls, *args, **kwargs):
        raise TypeError("Can't instantiate directly")


class FuncChecker(six.with_metaclass(ABCMeta, _NoInstances)):
    """函数相关的检查类"""

    @staticmethod
    def check_iscallable(func):
        """
        检查传入参数对象是否是函数；不是函数raise CheckError
        :param func: 传入参数对象
        """
        if not callable(func):
            raise CheckError('%s is not callable' % get_func_name(func))


class ArgChecker(six.with_metaclass(ABCMeta, _NoInstances)):
    """函数参数相关的检查类"""

    @staticmethod
    def check_type(*ty_args, **ty_kwargs):
        """
        【装饰器】
        检查输入参数类型；检查失败raise CheckError
        :param ty_args: 类型tuple
        :param ty_kwargs: 类型dict
        :return: 
        """
        # 检查是否有不合规的tuple参数
        for ty in ty_args:
            if not isinstance(ty, (type, tuple)):
                raise TypeError(
                    "check_type() expected a type or tuple of types"
                    ", but got {type_} instead.".format(
                        type_=ty,
                    )
                )
        # 检查是否有不合规的dict参数
        for name, ty in six.iteritems(ty_kwargs):
            if not isinstance(ty, (type, tuple)):
                raise TypeError(
                    "check_type() expected a type or tuple of types for "
                    "argument '{name}', but got {type_} instead.".format(
                        name=name, type_=ty,
                    )
                )
        # 将type_check作用在函数参数上
        return arg_process(*map(type_check, list(ty_args)), **valmap(type_check, ty_kwargs))

    @staticmethod
    def check_bound(*bd_args, **bd_kwargs):
        """
        【装饰器】
        检查输入参数是否在某一范围内；检查失败raise CheckError
        传入参数形式应为`` (min_value, max_value)``.
        ``None`` 可以作为 ``min_value`` 或 ``max_value``，相当于正负无穷
        :param bd_args: tuple范围参数
        :param bd_kwargs: dict范围参数
        :return: 
        """
        # 将bound_valid_and_check作用在函数参数上
        return arg_process(*map(bound_valid_and_check, list(bd_args)),
                           **valmap(bound_valid_and_check, bd_kwargs))

    @staticmethod
    def _check_default(check_no, *names, **_unused):
        """
        【装饰器】
        检查函数参数是否有或者没有默认值；检查失败raise CheckError
        :param staticmethod: 检查有 or 没有
        :param names: 待检查参数的名称
        :param _unused: 屏蔽dict参数
        :return: 
        """
        # 屏蔽dict参数
        if _unused:
            raise TypeError("_check_default() doesn't accept dict processors")
        check_err_msg = '' if check_no else 'no '

        def decorate(func):
            # 获取默认参数字典
            arg_defaults = get_arg_defaults(func)
            for name in names:
                if name not in arg_defaults:
                    # 传入了并不存在的参数名
                    raise TypeError(get_func_name(func) + ' has no argument named ' + name)
                if bool(check_no) != (isinstance(arg_defaults[name], ArgNoDefault)):
                    # 检查失败
                    raise CheckError(
                        'In ' + get_func_name(func) + ' argument {} has {}default'.format(name, check_err_msg))

            @wraps(func)
            def wrapper(*args, **kwargs):
                # 直接返回被装饰函数结果
                return func(*args, **kwargs)

            return wrapper

        return decorate

    @staticmethod
    def check_hasdefault(*names, **_unused):
        """
        【装饰器】
        检查函数参数是否有默认值；检查失败raise CheckError
        :param names: 待检查参数的名称
        :param _unused: 屏蔽dict参数
        :return: 
        """

        return ArgChecker._check_default(False, *names, **_unused)

    @staticmethod
    def check_nodefault(*names, **_unused):
        """
        【装饰器】
        检查函数参数是否没有默认值；检查失败raise CheckError
        :param names: 待检查参数的名称
        :param _unused: 屏蔽dict参数
        :return: 
        """

        return ArgChecker._check_default(True, *names, **_unused)

    @staticmethod
    def check_hasargs(func):
        """
        【装饰器】
        检查函数是否有*args参数；检查失败raise CheckError
        :param func: 传入函数对象
        :return: 
        """
        # 解包函数参数及默认值
        argspec = getargspec(func)
        spec_args = argspec.args if argspec.args else []
        defaults = argspec.defaults if argspec.defaults else ()
        if len(spec_args) - len(defaults) == 0:
            # 函数没有tuple参数
            raise CheckError(get_func_name(func) + ' has no args')

        @wraps(func)
        def wrapper(*args_inner, **kwargs):
            # 直接返回被装饰函数结果
            return func(*args_inner, **kwargs)

        return wrapper

    @staticmethod
    def check_haskwargs(func):
        """
        【装饰器
        检查函数是否有**kwargs参数；检查失败raise CheckError
        :param func: 传入函数对象
        :return: 
        """
        # 解包函数参数及默认值
        argspec = getargspec(func)
        defaults = argspec.defaults if argspec.defaults else ()
        if not defaults:
            # 函数没有dict参数
            raise CheckError(get_func_name(func) + ' has no kwargs')

        @wraps(func)
        def wrapper(*args, **kwargs):
            # 直接返回被装饰函数结果
            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def check_subset(*ss_args, **ss_kwargs):
        """
        【装饰器】
        检查输入参数是否是某一集合的子集；检查失败raise CheckError
        :param ss_args: 参数集合tuple
        :param ss_kwargs: 参数集合dict
        :return: 
        """
        # 检查是否有不合规的tuple参数
        for ss in ss_args:
            if not isinstance(ss, (list, set, type(None))):
                raise TypeError(
                    "check_subset() expected a list or set or None of values"
                    ", but got {subset_} or tuple instead.".format(
                        subset_=str(type(ss)),
                    )
                )
        # 检查是否有不合规的dict参数
        for name, ss in six.iteritems(ss_kwargs):
            if not isinstance(ss, (list, set, type(None))):
                raise TypeError(
                    "check_subset() expected a list or set of values for "
                    "argument '{name_}', but got {subset_} or tuple instead.".format(
                        name_=name, subset_=str(type(ss)),
                    )
                )
        # 将subset_check函数作用在函数参数上
        return arg_process(*map(subset_check, list(ss_args)), **valmap(subset_check, ss_kwargs))


class ReturnChecker(six.with_metaclass(ABCMeta, _NoInstances)):
    """函数返回值相关的检查类"""

    @staticmethod
    def check_type(*types, **_unused):
        """
        【装饰器】
        检查返回值类型；检查失败raise CheckError
        :param types: 类型tuple
        :param _unused: 屏蔽dict参数
        :return: 
        """
        # 屏蔽dict参数
        if _unused:
            raise TypeError("check_type() doesn't accept dict processors")

        # 检查是否有不合格的tuple参数
        for type_ in types:
            if not isinstance(type_, (type, tuple, type(None))):  # tuple or 内置类型
                raise TypeError(
                    "check_return_type() expected a type or tuple of types, but got {type_msg} instead.".format(
                        type_msg=type_,
                    )
                )
        # 将type_check函数作用在函数返回值上
        return return_process(*map(type_check, list(types)))

    @staticmethod
    def check_bound(*bounds, **_unused):
        """
        【装饰器】
        检查返回参数是否在某一范围内；检查失败raise CheckError
        传入参数形式应为`` (min_value, max_value)``.
        ``None`` 可以作为 ``min_value`` 或 ``max_value``，相当于正负无穷
        :param bounds: tuple范围参数
        :param _unused: 屏蔽dict参数
        :return: 
        """
        # 屏蔽dict参数
        if _unused:
            raise TypeError("check_bound() doesn't accept dict processors")
        # 将bound_valid_and_check函数作用在函数返回值上
        return return_process(*map(bound_valid_and_check, list(bounds)))

    @staticmethod
    def check_subset(*ss_args, **_unused):
        """
        【装饰器】
        检查输入参数是否是某一集合的子集；检查失败raise CheckError
        :param ss_args: 参数集合tuple
        :param _unused: 屏蔽dict参数
        :return: 
        """
        # 屏蔽dict参数
        if _unused:
            raise TypeError("check_subset() doesn't accept dict processors")
        # 检查传入的tuple参数
        for ss in ss_args:
            if not isinstance(ss, (list, set, type(None))):
                raise TypeError(
                    "check_subset() expected a list or set or None of values"
                    ", but got {subset_} or tuple instead.".format(
                        subset_=str(type(ss)),
                    )
                )
        # 将subset_check函数作用在函数返回值上
        return return_process(*map(subset_check, list(ss_args)))



from __future__ import absolute_import

from . import ABuDTUtil
from . import ABuDelegateUtil
from . import ABuDateUtil
from . import ABuFileUtil
from . import ABuMd5
from . import ABuRegUtil
from . import ABuScalerUtil
from . import ABuStatsUtil
from . import ABuStrUtil
from . import ABuProgress
from . import ABuPlatform
from . import ABuKLUtil
from . import ABuOsUtil

from .ABuProgress import AbuProgress, AbuBlockProgress, AbuMulPidProgress

__all__ = [
    'ABuDateUtil',
    'ABuDelegateUtil',
    'ABuDTUtil',
    'ABuFileUtil',
    'ABuMd5',
    'ABuRegUtil',
    'ABuScalerUtil',
    'ABuStatsUtil',
    'ABuStrUtil',
    'ABuProgress',
    'AbuProgress',
    'AbuBlockProgress',
    'AbuMulPidProgress',
    'ABuPlatform',
    'ABuKLUtil',
    'ABuOsUtil'
]



# coding=utf-8
"""
    数据源模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

import numpy as np

from ..MarketBu.ABuDataBase import BaseMarket
from ..MarketBu.ABuDataFeed import BDApi, TXApi, NTApi, HBApi, SNUSApi, SNFuturesApi, SNFuturesGBApi
from .ABuSymbol import Symbol
from .ABuSymbol import code_to_symbol
from ..CoreBu import ABuEnv
from ..CoreBu.ABuFixes import six
from ..CoreBu.ABuEnv import EMarketDataSplitMode, EMarketDataFetchMode
from ..CoreBu.ABuEnv import EMarketSourceType
from ..MarketBu.ABuDataCache import load_kline_df, load_kline_df_net
from ..UtilBu import ABuDateUtil

try:
    from tables import HDF5ExtError
except ImportError:
    class HDF5ExtError(RuntimeError):
        """如果没有HDF5环境只能使用其它存贮模式"""
        pass


"""内置数据源source_dict"""
source_dict = {EMarketSourceType.E_MARKET_SOURCE_bd.value: BDApi,
               EMarketSourceType.E_MARKET_SOURCE_tx.value: TXApi,
               EMarketSourceType.E_MARKET_SOURCE_nt.value: NTApi,
               EMarketSourceType.E_MARKET_SOURCE_sn_us.value: SNUSApi,
               EMarketSourceType.E_MARKET_SOURCE_sn_futures.value: SNFuturesApi,
               EMarketSourceType.E_MARKET_SOURCE_sn_futures_gb.value: SNFuturesGBApi,
               EMarketSourceType.E_MARKET_SOURCE_hb_tc.value: HBApi}


def _calc_start_end_date(df, force_local, n_folds, start, end):
    """
    根据参数计算start，end
    :param df: 本地缓存的金融时间序列对象，pd.DataFrame对象
    :param force_local: 是否强制走本地数据
    :param n_folds: 需要几年的数据
    :param start: 开始的时间
    :param end: 结束的时间
    :return:
    """

    # 当前今天时间日期str对象，如果是强制本地，即缓存的最后一个交易日
    today = ABuDateUtil.timestamp_to_str(df.index[-1]) if force_local else ABuDateUtil.current_str_date()
    if end is None:
        # 没有end也没start，end＝today，否则使用n_folds计算end
        end = today if start is None else ABuDateUtil.begin_date(-365 * n_folds, date_str=start, fix=False)
    # int类型的end, today转换
    end_int = ABuDateUtil.date_str_to_int(end)
    today_int = ABuDateUtil.date_str_to_int(today)
    if end_int > today_int:
        end_int = today_int

    if start is None:
        if force_local:
            end_ss = df[df.date >= end_int]
            if end_ss is None or end_ss.empty:
                ind = 0
            else:
                # +1补上
                ind = end_ss.key.values[0] - (ABuEnv.g_market_trade_year * n_folds) + 1
            if ind < 0:
                ind = 0
            # 强制本地，计算从第几个ind开始取df数据
            start = ABuDateUtil.timestamp_to_str(df.index[ind])
        else:
            # 非强制本地，使用n_folds年数，向前推计算start
            start = ABuDateUtil.begin_date(365 * n_folds, date_str=end, fix=False)
    start_int = ABuDateUtil.date_str_to_int(start)

    df_end_int = 0
    # 给正无穷匹配之后的start_int >= df_start_int
    df_start_int = np.inf

    if df is not None and df.shape[0] > 0:
        # 获取本地缓存df的开始，结束，字符串对象以及int值
        try:
            df_end = ABuDateUtil.timestamp_to_str(df.index[-1])
            df_end_int = ABuDateUtil.date_str_to_int(df_end)

            df_start = ABuDateUtil.timestamp_to_str(df.index[0])
            df_start_int = ABuDateUtil.date_str_to_int(df_start)
        except Exception as e:
            logging.exception(e)

    return end, end_int, df_end_int, start, start_int, df_start_int


def kline_pd(symbol, data_mode, n_folds=2, start=None, end=None, save=True):
    """
    统一调度选择内部或者外部数据源，决策是否本地数据读取，还是网络数据读取，以及根据不
    同的数据获取模式，调整数据的选择范围

    eg: n_fold=2, start=None, end=None ，从今天起往前数两年
        n_fold=2, start='2015-02-14', end=None， 从2015-02-14到现在，n_fold无效
        n_fold=2, start=None, end='2016-02-14'，从2016-02-14起往前数两年
        n_fold=2, start='2015-02-14', end='2016-02-14'，从start到end

    :param data_mode: EMarketDataSplitMode enum对象
    :param symbol: string or Symbol对象
                   e.g. 'sz300104'
                   e.g. Symbol(MType.SZ, '300104')
    :param n_folds: 年, 如果start不为空，则n_fold失效
    :param start: 开始时间 start为None时，start会根据end和n_fold计算出来，str对象
    :param end: 结束时间，str对象
    :param save: 从网络下载后是否缓存到本地
    """
    try:
        if isinstance(symbol, Symbol):
            temp_symbol = symbol
        elif isinstance(symbol, six.string_types):
            # 如果是str对象，通过code_to_symbol转化为Symbol对象
            temp_symbol = code_to_symbol(symbol)
        else:
            raise TypeError('symbol must like as "usTSLA" or "TSLA" or Symbol(MType.US, "TSLA")')
        if ABuEnv.g_private_data_source is None:
            # 如果没有设置私有数据源，使用env中设置的内置示例测试源
            source = source_dict[ABuEnv.g_market_source.value]
        else:
            # 有设置私有数据源
            source = ABuEnv.g_private_data_source
            # 私有源首先设置的需要是class类型，然后判断是BaseMarket的子类
            if not isinstance(source, six.class_types):
                raise TypeError('g_private_data_source must be a class type!!!')
            if not issubclass(ABuEnv.g_private_data_source, BaseMarket):
                raise TypeError('g_private_data_source must be a subclass of BaseMarket!!!')
        temp_symbol.source = source
        # 如果外部负责保存，就需要save_kl_key中相关信息
        save_kl_key = (temp_symbol, None, None)

        # symbol本地的pd.DataFrame数据缓存
        df = None
        # 本地的pd.DataFrame金融时间序列的第一个日期 int类型
        df_req_start = 0
        # 本地的pd.DataFrame金融时间序列的最后一个个日期 int类型
        df_req_end = 0

        if ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET:
            # 如果env中设置并非强制从网络获取数据，就从本地数据尝试读取df, df_req_start
            df, df_req_start, df_req_end = load_kline_df(temp_symbol.value)
        # 确定env中设置是否强制从本地缓存读取数据
        force_local = (ABuEnv.g_data_fetch_mode == EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL)

        if force_local and df is None:
            # 如果强制本地且df是空，直接返回
            return df, save_kl_key

        if data_mode == EMarketDataSplitMode.E_DATA_SPLIT_UNDO and force_local:
            # 如果强制本地且数据模式为E_DATA_SPLIT_UNDO，即不依据参数切割df，直接返回
            return df, save_kl_key

        # 标准化输入的start时间，eg 2016-7-26 －> 2016-07-26
        start = ABuDateUtil.fix_date(start)
        # 标准化输入的end时间，eg 2016-7-26 －> 2016-07-26
        end = ABuDateUtil.fix_date(end)
        # 根据n_folds，start，end计算需要请求的start，end
        end, end_int, df_end_int, start, start_int, df_start_int = _calc_start_end_date(df, force_local, n_folds, start,
                                                                                        end)
        save_kl_key = (temp_symbol, start_int, end_int)
        if ABuEnv.g_data_fetch_mode == EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET:
            # 如果是强制走网络，直接请求使用load_kline_df_net
            return load_kline_df_net(source, temp_symbol, n_folds=n_folds, start=start, end=end, start_int=start_int,
                                     end_int=end_int, save=save), save_kl_key

        # 检测本地缓存数据是否满足需要，如果需要的数据在存储的数据之间，则可切片放回
        match = False
        if start_int >= df_start_int and end_int <= df_end_int:
            match = True
        elif start_int >= df_start_int and force_local:
            match = True
        elif start_int >= df_req_start and end_int <= df_req_end:
            match = True

        if match:
            if data_mode == EMarketDataSplitMode.E_DATA_SPLIT_SE:
                # 如果满足，且模式需要根据切割df的进行切割筛选
                df = df[(start_int <= df.date) & (df.date <= end_int)]
        elif not force_local:
            # 数据不满足，但非强制本地，走网络
            df = load_kline_df_net(source, temp_symbol, n_folds, start=start, end=end, start_int=start_int,
                                   end_int=end_int, save=save)
            if data_mode == EMarketDataSplitMode.E_DATA_SPLIT_UNDO:
                # SPLIT_UNDO需要读取所有本地数据不切割返回
                df, _, _ = load_kline_df(temp_symbol.value)
        return df, save_kl_key
    except HDF5ExtError:
        # hdf5 bug
        logging.debug('{} HDF5ExtError'.format(symbol))
    except Exception as e:
        logging.info('Exception kline_pd symbol:{} e:{}'.format(symbol, e))
    return None, None



from __future__ import absolute_import

from .ABuFactorSellBase import AbuFactorSellBase, AbuFactorSellXD, ESupportDirection
from .ABuFactorPreAtrNStop import AbuFactorPreAtrNStop
from .ABuFactorAtrNStop import AbuFactorAtrNStop
from .ABuFactorCloseAtrNStop import AbuFactorCloseAtrNStop
from .ABuFactorSellBreak import AbuFactorSellBreak
from .ABuFactorSellNDay import AbuFactorSellNDay
from .ABuFactorSellDM import AbuDoubleMaSell

# noinspection all
from . import ABuFS as fs

__all__ = [
    'AbuFactorSellBase',
    'AbuFactorSellXD',
    'ESupportDirection',
    'AbuFactorPreAtrNStop',
    'AbuFactorAtrNStop',
    'AbuFactorCloseAtrNStop',
    'AbuFactorSellBreak',
    'AbuFactorSellNDay',
    'AbuDoubleMaSell',
    'fs'
]



# -*- encoding:utf-8 -*-
"""量化相关分析工具图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
from contextlib import contextmanager

import ipywidgets as widgets
from IPython.display import display

from ..WidgetBu.ABuWGToolBase import WidgetToolBase, multi_fetch_symbol_df_analyse
from ..WidgetBu.ABuWGBase import browser_down_csv_zip
from ..UtilBu import ABuProgress, ABuFileUtil
from ..MarketBu.ABuSymbol import code_to_symbol
from ..SimilarBu.ABuCorrcoef import ECoreCorrType, corr_matrix
from ..SimilarBu.ABuSimilar import find_similar_with_se, find_similar_with_folds
from ..TLineBu.ABuTLSimilar import calc_similar, coint_similar
from ..CoreBu.ABuEnv import EMarketTargetType, EDataCacheType
from ..CoreBu import ABuEnv
from ..UtilBu.ABuStatsUtil import cosine_distance_matrix, manhattan_distance_matrix, euclidean_distance_matrix

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnusedLocal,PyProtectedMember
class WidgetSMTool(WidgetToolBase):
    """数据分析界面"""

    def __init__(self, tool_set):
        """初始化相关分析界面"""
        super(WidgetSMTool, self).__init__(tool_set)

        corr = self.init_corr_ui()
        distances = self.init_distances_ui()
        market_corr = self.init_market_corr_ui()
        relative_corr = self.init_relative_corr_ui()
        coint_corr = self.init_coint_corr_ui()

        children = [corr, distances, market_corr, relative_corr, coint_corr]
        if self.scroll_factor_box:
            sm_box = widgets.Box(children,
                                 layout=self.scroll_widget_layout)
            # 需要再套一层VBox，不然外部的tab显示有问题
            self.widget = widgets.VBox([sm_box])
        else:
            # 一行显示两个，2个为一组，组装sub_children_group序列,
            sub_children_group = self._sub_children(children, len(children) / 2)
            sub_children_box = [widgets.HBox(sub_children) for sub_children in sub_children_group]
            self.widget = widgets.VBox(sub_children_box)

    @contextmanager
    def _init_widget_list_action(self, callback_analyse, analyse_name, n_target):
        """上下文管理器，上文封装tip label，下文封装action按钮，widget_list做连接"""
        if not callable(callback_analyse):
            raise TabError('callback_analyse must callable!')
        widget_list = []
        tip_label = widgets.Label(self.map_tip_target_label(n_target=n_target), layout=self.label_layout)
        widget_list.append(tip_label)

        yield widget_list

        analyse_bt = widgets.Button(description=analyse_name, layout=widgets.Layout(width='98%'),
                                    button_style='info')
        analyse_bt.on_click(callback_analyse)
        widget_list.append(analyse_bt)

    def init_coint_corr_ui(self):
        """全市场协整相关分析ui"""
        with self._init_widget_list_action(self.coint_corr_market_analyse, u'全市场协整相关分析', 1) as widget_list:
            coint_similar_description = widgets.Textarea(
                value=u'全市场协整相关分析: \n'
                      u'综合利用相关和协整的特性返回查询的股票是否有统计套利的交易机会\n'
                      u'1. 通过相关性分析筛选出与查询股票最相关的前100支股票作为种子\n'
                      u'2. 从种子中通过计算协整程度来度量查询股票是否存在统计套利机会\n'
                      u'3. 可视化整个过程\n',
                disabled=False,
                layout=self.description_layout
            )
            widget_list.append(coint_similar_description)

            tip_label1 = widgets.Label(u'全市场相对相关分析不支持实时网络数据模式', layout=self.label_layout)
            tip_label2 = widgets.Label(u'非沙盒模式需先用\'数据下载界面操作\'进行下载', layout=self.label_layout)
            self.coint_corr_data_mode = widgets.RadioButtons(
                options={u'沙盒数据模式': True, u'本地数据模式': False},
                value=True,
                description=u'数据模式:',
                disabled=False
            )
            corr_market_box = widgets.VBox([tip_label1, tip_label2, self.coint_corr_data_mode])
            widget_list.append(corr_market_box)
        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    def init_relative_corr_ui(self):
        """全市场相对相关分析ui"""

        with self._init_widget_list_action(self.corr_relative_market_analyse, u'全市场相对相关分析', 2) as widget_list:
            relative_description = widgets.Textarea(
                value=u'全市场相对相关分析: \n'
                      u'度量的是两目标（a,b）相对整个市场的相关性评级，它不关心某一个股票具体相关性的数值的大小\n'
                      u'1. 计算a与市场中所有股票的相关性\n'
                      u'2. 将所有相关性进行rank排序\n'
                      u'3. 查询股票b在rank序列中的位置，此位置值即为结果\n',
                disabled=False,
                layout=self.description_layout
            )
            widget_list.append(relative_description)

            tip_label1 = widgets.Label(u'全市场相对相关分析不支持实时网络数据模式', layout=self.label_layout)
            tip_label2 = widgets.Label(u'非沙盒模式需先用\'数据下载界面操作\'进行下载', layout=self.label_layout)
            self.relative_corr_data_mode = widgets.RadioButtons(
                options={u'沙盒数据模式': True, u'本地数据模式': False},
                value=True,
                description=u'数据模式:',
                disabled=False
            )
            corr_market_box = widgets.VBox([tip_label1, tip_label2, self.relative_corr_data_mode])
            widget_list.append(corr_market_box)

        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    def init_market_corr_ui(self):
        """全市场相关分析ui"""

        with self._init_widget_list_action(self.corr_market_analyse, u'全市场相关分析', 1) as widget_list:
            self.corr_market = widgets.Dropdown(
                options={u'美股': EMarketTargetType.E_MARKET_TARGET_US.value,
                         u'A股': EMarketTargetType.E_MARKET_TARGET_CN.value,
                         u'港股': EMarketTargetType.E_MARKET_TARGET_HK.value,
                         u'国内期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_CN.value,
                         u'国际期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL.value,
                         u'数字货币': EMarketTargetType.E_MARKET_TARGET_TC.value},
                value=ABuEnv.g_market_target.value,
                description=u'分析市场:',
            )
            market_tip1 = widgets.Label(value=u'分析市场的选择可以和分析目标不在同一市场', layout=self.label_layout)
            market_tip2 = widgets.Label(value=u'如分析目标为美股股票，分析市场也可选A股', layout=self.label_layout)
            market_box = widgets.VBox([market_tip1, market_tip2, self.corr_market])
            widget_list.append(market_box)

            tip_label1 = widgets.Label(u'全市场相关分析不支持开放数据模式', layout=self.label_layout)
            tip_label2 = widgets.Label(u'非沙盒模式需先用\'数据下载界面操作\'进行下载', layout=self.label_layout)
            self.corr_market_data_mode = widgets.RadioButtons(
                options={u'沙盒数据模式': True, u'本地数据模式': False},
                value=True,
                description=u'数据模式:',
                disabled=False
            )
            corr_market_box = widgets.VBox([tip_label1, tip_label2, self.corr_market_data_mode])
            widget_list.append(corr_market_box)

            self.corr_market_mode = widgets.RadioButtons(
                options={u'皮尔逊相关系数计算': 'pears', u'斯皮尔曼相关系数计算': 'sperm', u'基于＋－符号相关系数': 'sign',
                         u'移动时间加权相关系数': 'rolling'},
                value='pears',
                description=u'相关模式:',
                disabled=False
            )
            widget_list.append(self.corr_market_mode)

        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    def init_corr_ui(self):
        """相关分析ui"""

        with self._init_widget_list_action(self.corr_analyse, u'相关分析', -1) as widget_list:
            self.corr_mode = widgets.RadioButtons(
                options={u'皮尔逊相关系数计算': 'pears', u'斯皮尔曼相关系数计算': 'sperm', u'基于＋－符号相关系数': 'sign',
                         u'移动时间加权相关系数': 'rolling'},
                value='pears',
                description=u'相关模式:',
                disabled=False
            )
            widget_list.append(self.corr_mode)

        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    def init_distances_ui(self):
        """距离分析ui"""

        with self._init_widget_list_action(self.distances_analyse, u'距离分析', -1) as widget_list:
            self.distances_mode = widgets.RadioButtons(
                options={u'曼哈顿距离(L1范数)': 0, u'欧式距离(L2范数)': 1, u'余弦距离': 2},
                value=0,
                description=u'距离模式:',
                disabled=False
            )
            widget_list.append(self.distances_mode)

            scale_end_label = widgets.Label(u'对结果矩阵进行标准化处理', layout=self.label_layout)
            self.scale_end = widgets.Checkbox(
                value=True,
                description=u'标准化',
                disabled=False
            )
            scale_end_box = widgets.VBox([scale_end_label, self.scale_end])
            widget_list.append(scale_end_box)

            similar_tip_label = widgets.Label(u'对结果矩阵进行转换相关性', layout=self.label_layout)
            self.to_similar = widgets.Checkbox(
                value=False,
                description=u'转换相关',
                disabled=False
            )
            to_similar_box = widgets.VBox([similar_tip_label, self.to_similar])
            widget_list.append(to_similar_box)

        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    @multi_fetch_symbol_df_analyse('p_change')
    def corr_analyse(self, cg_df, bt):
        """通过corr_matrix进行矩阵相关分析action"""
        display(corr_matrix(cg_df, similar_type=ECoreCorrType(self.corr_mode.value)))
        cg_df.cumsum().plot()

    @multi_fetch_symbol_df_analyse('p_change')
    def distances_analyse(self, cg_df, bt):
        """通过l1, l2, 余弦距离进行距离分析action"""
        if self.distances_mode.value == 0:
            distance = manhattan_distance_matrix(cg_df, scale_end=self.scale_end.value,
                                                 to_similar=self.to_similar.value)
        elif self.distances_mode.value == 1:
            distance = euclidean_distance_matrix(cg_df, scale_end=self.scale_end.value,
                                                 to_similar=self.to_similar.value)
        else:
            distance = cosine_distance_matrix(cg_df, scale_end=self.scale_end.value,
                                              to_similar=self.to_similar.value)
        display(distance)
        cg_df.cumsum().plot()

    def coint_corr_market_analyse(self, bt):
        """全市场协整相关分析"""
        with self.data_mode_recover(self.coint_corr_data_mode.value):
            ABuProgress.clear_output()
            if not self.all_market_env_check():
                return

            symbol = self._choice_symbol_single(default='usAAPL')

            mk = '{}_{}_sum_rank'.format(code_to_symbol(symbol).market.value, ABuEnv._g_enable_example_env_ipython)
            sum_rank = getattr(self, mk, None)
            _, sum_rank = coint_similar(symbol, sum_rank=sum_rank, show=True)
            # 缓存sum_rank
            setattr(self, mk, sum_rank)

    def corr_relative_market_analyse(self, bt):
        """全市场相对相关分析action"""
        with self.data_mode_recover(self.relative_corr_data_mode.value):
            ABuProgress.clear_output()
            if not self.all_market_env_check():
                return

            symbol1, symbol2 = self._choice_symbol_pair(default=['sh600036', 'sh601766'])
            mk = '{}_{}_sum_rank'.format(code_to_symbol(symbol1).market.value, ABuEnv._g_enable_example_env_ipython)
            sum_rank = getattr(self, mk, None)
            _, sum_rank = calc_similar(symbol1, symbol2, sum_rank=sum_rank, show=True)
            # 缓存sum_rank
            setattr(self, mk, sum_rank)

    def corr_market_analyse(self, bt):
        """全市场相关分析"""
        with self.data_mode_recover(self.corr_market_data_mode.value):
            ABuProgress.clear_output()

            if not self.all_market_env_check():
                return

            # 不做全局设置，做为后还需要恢复
            tmp_market = ABuEnv.g_market_target
            ABuEnv.g_market_target = EMarketTargetType(self.corr_market.value)

            symbol = self._choice_symbol_single(default='usAAPL')
            start, end, n_folds = self._start_end_n_fold()
            corr_type = ECoreCorrType(self.corr_market_mode.value)
            if start is not None and end is not None:
                find_similar_with_se(symbol, start=start, end=end, corr_type=corr_type)
            else:
                find_similar_with_folds(symbol, n_folds=n_folds, corr_type=corr_type)

            ABuEnv.g_market_target = tmp_market

    def all_market_env_check(self):
        """如果要做全市场相关类型的操作，检测本地数据文件"""
        if ABuEnv._g_enable_example_env_ipython:
            # 沙盒环境下不需要检测
            return True

        if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_CSV:
            # csv模式下
            if not ABuFileUtil.file_exist(ABuEnv.g_project_kl_df_data_csv):
                # 全市场回测，但没有数据
                logging.info(
                    u'全市场相关操作为了提高运行效率, 只使用\'本地数据模式\'进行回测，但未发现本地缓存数据，'
                    u'如需要进行数据更新'
                    u'请先使用\'数据下载界面操作\'进行数据更新！')
                browser_down_csv_zip()
                return False
            elif len(os.listdir(ABuEnv.g_project_kl_df_data_csv)) < 30:
                # 全市场回测，但数据不足, 这里取30
                logging.info(
                    u'全市场相关操作为了提高运行效率, 只使用\'本地数据模式\'进行回测，发现本地缓存数据不足，'
                    u'只有{}支股票历史数据信息'
                    u'如需要进行数据更新'
                    u'请先使用\'数据下载界面操作\'进行数据更新！'.format(
                        len(os.listdir(ABuEnv.g_project_kl_df_data_csv))))
                browser_down_csv_zip()
                return False
        elif ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_HDF5 \
                and not ABuFileUtil.file_exist(ABuEnv.g_project_kl_df_data):
            # hdf5模式下文件不存在
            logging.info(
                u'全市场相关为了提高运行效率, 只使用\'本地数据模式\'进行回测'
                u'hdf5模式下未发现hdf5本地缓存数据，'
                u'如需要进行数据更新'
                u'请先使用\'数据下载界面操作\'进行数据更新！')
            browser_down_csv_zip()
            return False
        return True



# -*- encoding:utf-8 -*-
"""
    交易可视化模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import copy

import logging
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ..CoreBu import ABuEnv
from ..UtilBu import ABuDateUtil
from ..UtilBu.ABuProgress import AbuProgress

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import range
from ..TradeBu.ABuCapital import AbuCapital

g_enable_his_corr = True
g_enable_his_trade = True

__author__ = '阿布'
__weixin__ = 'abu_quant'


def plot_his_trade(orders, kl_pd):
    """
    可视化绘制AbuOrder对象，绘制交易买入时间，卖出时间，价格，生效因子等
    :param orders: AbuOrder对象序列
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :return:
    """

    if not g_enable_his_trade:
        return

    # 拿出时间序列中最后一个，做为当前价格
    now_price = kl_pd.iloc[-1].close
    all_pd = kl_pd

    # ipython环境绘制在多个子画布上，普通python环境绘制一个show一个
    draw_multi_ax = ABuEnv.g_is_ipython

    # 根据绘制环境设置子画布数量
    ax_cnt = 1 if not draw_multi_ax else len(orders)
    # 根据子画布数量设置画布大小
    plt.figure(figsize=(14, 8 * ax_cnt))
    fig_dims = (ax_cnt, 1)

    with AbuProgress(len(orders), 0) as pg:
        for index, order in enumerate(orders):
            pg.show(index + 1)
            # 迭代所有orders，对每一个AbuOrder对象绘制交易细节
            mask_date = all_pd['date'] == order.buy_date
            st_key = all_pd[mask_date]['key']

            if order.sell_type == 'keep':
                rv_pd = all_pd.iloc[st_key.values[0]:, :]
            else:
                mask_sell_date = all_pd['date'] == order.sell_date
                st_sell_key = all_pd[mask_sell_date]['key']
                rv_pd = all_pd.iloc[st_key.values[0]:st_sell_key.values[0], :]

            if draw_multi_ax:
                # ipython环境绘制在多个子画布上
                plt.subplot2grid(fig_dims, (index, 0))
            # 绘制价格曲线
            plt.plot(all_pd.index, all_pd['close'], label='close')
            # 填充透明blue
            plt.fill_between(all_pd.index, 0, all_pd['close'], color='blue', alpha=.18)

            if order.sell_type == 'keep':
                # 如果单子还没卖出，是否win使用now_price代替sell_price，需＊单子期望的盈利方向
                order_win = (now_price - order.buy_price) * order.expect_direction > 0
            elif order.sell_type == 'win':
                order_win = True
            else:
                order_win = False

            if order_win:
                # 盈利的使用红色
                plt.fill_between(rv_pd.index, 0, rv_pd['close'], color='red', alpha=.38)
            else:
                # 亏损的使用绿色
                plt.fill_between(rv_pd.index, 0, rv_pd['close'], color='green', alpha=.38)

            # 格式化买入信息标签
            buy_date_fmt = ABuDateUtil.str_to_datetime(str(order.buy_date), '%Y%m%d')
            buy_tip = 'buy_price:{:.2f}'.format(order.buy_price)

            # 写买入tip信息
            plt.annotate(buy_tip, xy=(buy_date_fmt, all_pd['close'].asof(buy_date_fmt) * 2 / 5),
                         xytext=(buy_date_fmt, all_pd['close'].asof(buy_date_fmt)),
                         arrowprops=dict(facecolor='red'),
                         horizontalalignment='left', verticalalignment='top')

            if order.sell_price is not None:
                # 如果单子卖出，卖出入信息标签使用，收益使用sell_price计算，需＊单子期望的盈利方向
                sell_date_fmt = ABuDateUtil.str_to_datetime(str(order.sell_date), '%Y%m%d')
                pft = (order.sell_price - order.buy_price) * order.buy_cnt * order.expect_direction
                sell_tip = 'sell price:{:.2f}, profit:{:.2f}'.format(order.sell_price, pft)
            else:
                # 如果单子未卖出，卖出入信息标签使用，收益使用now_price计算，需＊单子期望的盈利方向
                sell_date_fmt = ABuDateUtil.str_to_datetime(str(all_pd[-1:]['date'][0]), '%Y%m%d')
                pft = (now_price - order.buy_price) * order.buy_cnt * order.expect_direction
                sell_tip = 'now price:{:.2f}, profit:{:.2f}'.format(now_price, pft)

            # 写卖出tip信息
            plt.annotate(sell_tip, xy=(sell_date_fmt, all_pd['close'].asof(sell_date_fmt) * 2 / 5),
                         xytext=(sell_date_fmt, all_pd['close'].asof(sell_date_fmt)),
                         arrowprops=dict(facecolor='green'),
                         horizontalalignment='left', verticalalignment='top')
            # 写卖出因子信息
            plt.annotate(order.sell_type_extra, xy=(buy_date_fmt, all_pd['close'].asof(sell_date_fmt) / 4),
                         xytext=(buy_date_fmt, all_pd['close'].asof(sell_date_fmt) / 4),
                         arrowprops=dict(facecolor='yellow'),
                         horizontalalignment='left', verticalalignment='top')

            # 写买入因子信息
            if order.buy_factor is not None:
                plt.annotate(order.buy_factor, xy=(buy_date_fmt, all_pd['close'].asof(sell_date_fmt) / 3),
                             xytext=(buy_date_fmt, all_pd['close'].asof(sell_date_fmt) / 3),
                             arrowprops=dict(facecolor='yellow'),
                             horizontalalignment='left', verticalalignment='top')
            # title使用时间序列symbol
            plt.title(order.buy_symbol)
            if not draw_multi_ax:
                # ipython环境绘制在多个子画布上，普通python环境绘制一个show一个
                plt.show()

    plt.show()


def plot_capital_info(capital_pd, init_cash=-1):
    """
    资金信息可视化
    :param capital_pd: AbuCapital对象或者AbuCapital对象的capital_pd
    :param init_cash: 初始化cash，如果capital_pd为AbuCapital对象，即从capital_pd获取
    """

    if isinstance(capital_pd, AbuCapital):
        # 如果是AbuCapital对象进行转换
        init_cash = capital_pd.read_cash
        capital_pd = capital_pd.capital_pd

    if init_cash != -1:
        cb_earn = capital_pd['capital_blance'] - init_cash
        try:
            # 从有资金变化开始的loc开始绘制
            # noinspection PyUnresolvedReferences
            cb_earn = cb_earn.loc[cb_earn[cb_earn != 0].index[0]:]
            cb_earn.plot()
            plt.title('capital_blance earn from none zero point')
            plt.show()
            sns.regplot(x=np.arange(0, cb_earn.shape[0]), y=cb_earn.values, marker='+')
            plt.show()
        except Exception as e:
            logging.exception(e)
            capital_pd['capital_blance'].plot()
            plt.title('capital blance')
            plt.show()

    # 为了画出平滑的曲线，取有值的
    cap_cp = copy.deepcopy(capital_pd)
    cap_cp['stocks_blance'][cap_cp['stocks_blance'] <= 0] = np.nan
    cap_cp['stocks_blance'].fillna(method='pad', inplace=True)
    cap_cp['stocks_blance'].dropna(inplace=True)
    cap_cp['stocks_blance'].plot()
    plt.title('stocks blance')
    plt.show()

    try:
        sns.distplot(capital_pd['capital_blance'], kde_kws={"lw": 3, "label": "capital blance kde"})
        plt.show()
    except Exception as e:
        logging.debug(e)
        capital_pd['capital_blance'].plot(kind='kde')
        plt.title('capital blance kde')
        plt.show()


def plot_bk_xd(bk_summary, kl_pd_xd_mean, title=None):
    """根据有bk_summary属性的bk交易因子进行可视化，暂时未迁移完成"""
    plt.figure()
    plt.plot(list(range(0, len(kl_pd_xd_mean))), kl_pd_xd_mean['close'])
    for bk in bk_summary.bk_xd_obj_list:
        plt.hold(True)
        pc = 'r' if bk.break_sucess is True else 'g'
        plt.plot(bk.break_index, kl_pd_xd_mean['close'][bk.break_index], 'ro', markersize=12, markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor=pc)
    if title is not None:
        plt.title(title)
    plt.grid(True)


def plot_kp_xd(kp_summary, kl_pd_xd_mean, title=None):
    """根据有bk_summary属性的kp交易因子进行可视化，暂时未迁移完成"""
    plt.figure()
    plt.plot(list(range(0, len(kl_pd_xd_mean))), kl_pd_xd_mean['close'])

    for kp in kp_summary.kp_xd_obj_list:
        plt.hold(True)
        plt.plot(kp.break_index, kl_pd_xd_mean['close'][kp.break_index], 'ro', markersize=8, markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor='r')

    if title is not None:
        plt.title(title)
    plt.grid(True)



# IN ABu LIST __all__
from __future__ import absolute_import

from .ABuFixes import *
from .ABuPdHelper import *
from . import ABuEnv as env
from . import ABu as abu
from .ABuEnv import EMarketSourceType, EMarketTargetType, EMarketSubType, \
    EMarketDataSplitMode, EMarketDataFetchMode, EDataCacheType
from .ABuBase import AbuParamBase, FreezeAttrMixin, PickleStateMixin
from .ABuParallel import Parallel, delayed
from .ABuStore import AbuResultTuple, EStoreAbu

__all__ = [
    'abu',
    'AbuResultTuple',
    'EStoreAbu',
    'env',
    'AbuParamBase',
    'FreezeAttrMixin',
    'PickleStateMixin',
    'Parallel',
    'delayed',
    'EMarketSourceType',
    'EMarketTargetType',
    'EMarketSubType',
    'EMarketDataSplitMode',
    'EMarketDataFetchMode',
    'EDataCacheType',

    'train_test_split',
    'KFold',
    'learning_curve',
    'cross_val_score',
    'GridSearchCV',

    'signature',
    'Parameter',

    'ThreadPoolExecutor',
    'six',
    'zip',
    'xrange',
    'range',
    'reduce',
    'map',
    'filter',
    'pickle',
    'Pickler',
    'Unpickler',
    'partial',

    'pd_rolling_mean',
    'pd_rolling_std',
    'pd_rolling_var',
    'pd_rolling_median',
    'pd_rolling_max',
    'pd_rolling_min',
    'pd_rolling_corr',
    'pd_rolling_cov',
    'pd_rolling_sum',
    'pd_rolling_kurt',
    'pd_rolling_skew',

    'pd_ewm_mean',
    'pd_ewm_corr',
    'pd_ewm_std',
    'pd_ewm_cov',
    'pd_ewm_var',

    'pd_expanding_mean',
    'pd_expanding_std',
    'pd_expanding_var',
    'pd_expanding_median',
    'pd_expanding_max',
    'pd_expanding_min',
    'pd_expanding_corr',
    'pd_expanding_cov',
    'pd_expanding_sum',
    'pd_expanding_kurt',
    'pd_expanding_skew',

    'pd_resample'
]



# -*- encoding:utf-8 -*-
"""
    内置特征定义，以及用户特征扩展，定义模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import datetime
import os

import numpy as np

from ..CoreBu import ABuEnv
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange, range, six
from ..MarketBu import ABuMarketDrawing
from ..TLineBu import ABuTLAtr
from ..TLineBu import ABuTLJump
from ..TLineBu import ABuTLWave
from ..UtilBu import ABuRegUtil
from ..UtilBu import ABuStrUtil

__author__ = '阿布'
__weixin__ = 'abu_quant'

# 内置特征，趋势角度
g_deg_keys = [21, 42, 60, ABuEnv.g_market_trade_year]
# 内置特征，价格rank
g_price_rank_keys = [60, 90, 120, ABuEnv.g_market_trade_year]
# 内置特征，波动周期定义
g_wave_xd = 42
# 内置特征，波动取样个数
g_wave_key_cnt = 3
# 内置特征，atr周期定义
g_atr_xd = 42
# 快照周期
g_take_snap_shot_xd = 60


class BuyFeatureMixin(object):
    """
        买入特征标识混入，与BuyUmpMixin不同，具体feature类可能属于多个类别
        即可能同时混入BuyFeatureMixin和SellFeatureMixin
    """
    _feature_buy = True
    _feature_buy_prefix = 'buy_'


class SellFeatureMixin(object):
    """
        卖出特征标识混入，与SellUmpMixin不同，具体feature类可能属于多个类别
        即可能同时混入BuyFeatureMixin和SellFeatureMixin
    """
    _feature_sell = True
    _feature_sell_prefix = 'sell_'


class AbuFeatureBase(object):
    """特征构造基类"""

    def support_buy_feature(self):
        """是否支持买入特征构建"""
        return getattr(self, "_feature_buy", False) is True

    def support_sell_feature(self):
        """是否支持卖出特征构建"""
        return getattr(self, "_feature_sell", False) is True

    def check_support(self, buy_feature):
        """
        根据参数buy_feature检测是否支持特征构建
        :param buy_feature: 是否是买入特征构造（bool）
        """
        if buy_feature and not self.support_buy_feature:
            raise TypeError('feature support buy must subclass BuyFeatureMixin!!!')
        if not buy_feature and not self.support_sell_feature:
            raise TypeError('feature support buy must subclass SellFeatureMixin!!!')

    def feature_prefix(self, buy_feature, check=True):
        """
        根据buy_feature决定返回_feature_buy_prefix或者_feature_sell_prefix，目的是在calc_feature中构成唯一key
        :param buy_feature: 是否是买入特征构造（bool）
        :param check: 是否需要检测是否支持特征构建
        :return:
        """
        if check:
            self.check_support(buy_feature)
        return getattr(self, '_feature_buy_prefix') if buy_feature else getattr(self, '_feature_sell_prefix')

    def __str__(self):
        """打印对象显示：class name, support_buy_feature support_sell_feature, get_feature_keys"""
        return '{}:is_buy_feature:{} is_sell_feature:{} feature: {}'.format(self.__class__.__name__,
                                                                            self.support_buy_feature(),
                                                                            self.support_sell_feature(),
                                                                            self.get_feature_keys(
                                                                                self.support_buy_feature()))

    __repr__ = __str__

    def get_feature_ump_keys(self, ump_cls):
        """
        根据ump_cls，返回对应的get_feature_keys
        :param ump_cls: AbuUmpEdgeBase子类，参数为类，非实例对象
        :return: 键值对字典中的key序列
        """
        is_buy_ump = getattr(ump_cls, "_ump_type_prefix") == 'buy_'
        return self.get_feature_keys(buy_feature=is_buy_ump)

    def get_feature_keys(self, buy_feature):
        """
        子类主要需要实现的函数，定义feature的列名称
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 键值对字典中的key序列
        """
        raise NotImplementedError('NotImplementedError get_feature_keys!!!')

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        子类主要需要实现的函数，根据买入或者卖出时的金融时间序列，以及交易日信息构造特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造特征的键值对字典
        """
        raise NotImplementedError('NotImplementedError calc_feature!!!')


class AbuFeatureDeg(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """角度特征，支持买入，卖出"""

    def __init__(self):
        """
            默认21, 42, 60, 250日走势角度特征，如外部修改，直接使用类似如下：
                       abupy.feature.g_deg_keys = [10, 20, 30, 40, 50]
        """

        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.deg_keys = frozenset(g_deg_keys)

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有走势角度特征feature的列名称定, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 角度特征的键值对字典中的key序列
        """
        return ['{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature), dk) for dk in self.deg_keys]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造拟合角度特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造角度特征的键值对字典
        """

        # 返回的角度特征键值对字典
        deg_dict = {}
        for dk in self.deg_keys:
            # 迭代预设角度周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                deg_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                deg_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[0] > dk else combine_kl_pd.close

            # 使用截取特征交易周期收盘价格deg_close做为参数，通过calc_regress_deg计算趋势拟合角度
            ang = ABuRegUtil.calc_regress_deg(deg_close, show=False)
            # 标准化拟合角度值
            ang = 0 if np.isnan(ang) else round(ang, 3)
            # 角度特征键值对字典添加拟合角度周期key和对应的拟合角度值
            deg_dict['{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature), dk)] = ang
        return deg_dict


class AbuFeaturePrice(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """价格rank特征，支持买入，卖出"""

    def __init__(self):
        """
            默认60, 90, 120, 250日价格rank特征，如外部修改，直接使用类似如下：
                    abupy.feature.g_price_rank_keys = [10, 20, 30, 40，50]
        """

        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.price_rank_keys = frozenset(g_price_rank_keys)

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有价格rank特征feature的列名称, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 价格rank特征的键值字典中的key序列
        """
        return ['{}price_rank{}'.format(self.feature_prefix(buy_feature=buy_feature), dk) for dk in
                self.price_rank_keys]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造价格rank特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造价格rank特征的键值对字典
        """
        # 返回的价格rank特征键值对字典
        price_rank_dict = {}
        for dk in self.price_rank_keys:
            # 迭代预设价格rank周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                price_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                price_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[0] > dk else combine_kl_pd.close
            """
                price_close.rank()的结果是所有价格的排名

                ...................
                2016-07-20    256.0
                2016-07-21    200.0
                2016-07-22    214.0
                2016-07-25    266.0
                2016-07-26    239.0

                -> price_close.rank()[-1]的结果是买入时刻或者卖出时刻的排名

                239.0

                －> price_close.rank()[-1] / price_close.rank().shape[0] 的结果即为：
                买入时刻或者卖出时刻的排名在周期中的排名位置，值由0-1

                eg: price_close.rank()[-1] / price_close.rank().shape[0]
                -> 239.0 / 504 = 0.47420634920634919, 即代表买入或者卖出时价格在特征周期中的位置
            """
            price_rank = price_close.rank()[-1] / price_close.rank().shape[0]
            # 标准化价格rank值
            price_rank = 0 if np.isnan(price_rank) else round(price_rank, 3)
            # 价格rank特征键值对字典添加价格rank周期key和对应的价格rank值
            price_rank_dict['{}price_rank{}'.format(self.feature_prefix(buy_feature=buy_feature), dk)] = price_rank
        return price_rank_dict


class AbuFeatureWave(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """波动特征，支持买入，卖出"""

    def __init__(self):
        """
            默认42日做为波动计算周期，如外部修改，直接使用类似如下：
                abupy.feature.g_wave_xd = 21
        """
        self.wave_xd = g_wave_xd
        self.wave_key_cnt = g_wave_key_cnt

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有波动特征feature的列名称, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 波动特征的键值对字典中的key序列
        """
        return ['{}wave_score{}'.format(self.feature_prefix(buy_feature=buy_feature),
                                        xd_ind) for xd_ind in list(range(1, self.wave_key_cnt + 1))]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造波动特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造波动特征的键值对字典
        """

        # 构建波动特征周期固定，即一年的交易日数量
        wave_wide = ABuEnv.g_market_trade_year
        if day_ind - wave_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            wave_df = kl_pd[day_ind - wave_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-wave_wide:].close，否则取combine_kl_pd所有交易收盘价格
            wave_df = combine_kl_pd[-wave_wide:] if combine_kl_pd.shape[0] > wave_wide else combine_kl_pd

        # 返回的波动特征键值对字典
        wave_dict = {}
        for xd_ind in xrange(1, self.wave_key_cnt + 1):
            # wave_df固定为一年交易时间序列，xd是calc_wave_st内部计算rolling std的window值，详ABuTLWave.calc_wave_std
            wave = ABuTLWave.calc_wave_std(wave_df, xd=xd_ind * self.wave_xd, show=False)
            wave_score = wave.score
            # 标准化波动特征值
            wave_score = 0 if np.isnan(wave_score) else round(wave_score, 3)
            # 波动特征键值对字典添加波动特征key和对应的波动特征值
            wave_dict['{}wave_score{}'.format(self.feature_prefix(buy_feature=buy_feature), xd_ind)] = wave_score
        return wave_dict


class AbuFeatureAtr(AbuFeatureBase, BuyFeatureMixin):
    """atr特征，支持买入"""

    def __init__(self):
        """
            默认42日做为atr特征计算周期，如外部修改，直接使用类似如下：
                abupy.feature.g_atr_xd = 21
        """
        self.atr_xd = g_atr_xd
        self.atr_key = 'atr_std'

    def get_feature_keys(self, buy_feature):
        """
        返回对应的atr特征key值，虽然只有一个固定的atr key值，也返回序列，保持接口统一
        :param buy_feature: 是否是买入特征构造（bool）
        :return: atr特征的键值对字典中的key序列
        """
        return ['{}{}'.format(self.feature_prefix(buy_feature=buy_feature), self.atr_key)]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造atr特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造atr特征的键值对字典
        """

        # 构建atr特征周期固定，即一年的交易日数量
        atr_wide = ABuEnv.g_market_trade_year
        if day_ind - atr_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            atr_df = kl_pd[day_ind - atr_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-atr_wide:].close，否则取combine_kl_pd所有交易收盘价格
            atr_df = combine_kl_pd[-atr_wide:] if combine_kl_pd.shape[0] > atr_wide else combine_kl_pd

        # 返回的atr特征键值对字典
        atr_dict = {}
        # 计算atr特征，详见ABuTLAtr.calc_atr_std
        atr_std = ABuTLAtr.calc_atr_std(atr_df, xd=self.atr_xd, show=False)
        atr_score = atr_std.score
        # 标准化atr特征
        atr_score = 0 if np.isnan(atr_score) else round(atr_score, 3)
        # atr特征键值对字典添加atr特征key和对应的atr特征值
        atr_dict['{}{}'.format(self.feature_prefix(buy_feature=buy_feature), self.atr_key)] = atr_score
        return atr_dict


class AbuFeatureJump(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """跳空特征，支持买入，卖出"""

    def __init__(self):
        """
            跳空的特征key为:
            jump_down_power: 向下跳空的能量
            diff_down_days : 向下跳空距离买入或卖出的时间间隔
            jump_up_power  : 向上跳空的能量
            diff_up_days   : 向上跳空距离买入或卖出的时间间隔
        """
        self.jump_keys = frozenset(['jump_down_power', 'diff_down_days', 'jump_up_power', 'diff_up_days'])

    def get_feature_keys(self, buy_feature):
        """
        返回对应的跳空特征key值
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 跳空特征的键值对字典中的key序列
        """
        return ['{}{}'.format(self.feature_prefix(buy_feature=buy_feature), jk) for jk in self.jump_keys]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造跳空特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造跳空特征的键值对字典
        """

        # 构建跳空特征周期固定，即一年的交易日数量
        jump_wide = ABuEnv.g_market_trade_year
        if day_ind - jump_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            jump_df = kl_pd[day_ind - jump_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-jump_wide:].close，否则取combine_kl_pd所有交易收盘价格
            jump_df = combine_kl_pd[-jump_wide:] if combine_kl_pd.shape[0] > jump_wide else combine_kl_pd

        # 返回的跳空特征键值对字典
        jump_dict = {}

        diff_down_days = 0
        jump_down_power = 0
        diff_up_days = 0
        jump_up_power = 0
        key_prefix = self.feature_prefix(buy_feature=buy_feature)
        # 通过特征周期时间序列jump_df做为参数，计算跳空，返回jumps对象为pd.DataFrame对象，详见ABuTLJump.calc_jump
        jumps = ABuTLJump.calc_jump(jump_df, show=False)
        """
                jump形式如下所示：jump代表跳空方向1为向上，－1为向下，jump_power代表对应的跳空能量，即每一行
                记录了一次跳空发生的时间，价格变化，跳空能量等信息
                            jump	jump_power	close	date	p_change	pre_close
                2014-08-11	 1.0	1.006085	259.32	20140811.0	4.51	248.13
                2014-10-10	-1.0	1.628481	236.91	20141010.0	-7.82	257.01
                2015-01-14	-1.0	1.325337	192.69	20150114.0	-5.66	204.25
                2015-02-12	-1.0	1.422285	202.88	20150212.0	-4.66	212.80
                .............
        """
        if not jumps.empty:
            # 筛选出所有向下跳空的
            down_jumps = jumps[(jumps.jump == -1)]
            if down_jumps.shape[0] > 0:
                # 筛选最后一次向下跳空的情况
                last_down_jump = down_jumps.iloc[-1:]
                # 跳空能力 * last_down_jump.jump.values[0] 转换成有方向的
                jump_down_power = last_down_jump.jump_power.values[0] * last_down_jump.jump.values[0]
                # 向下跳空距离买入或卖出的时间间隔
                diff_down_days = (kl_pd.iloc[day_ind: day_ind + 1].index.date - last_down_jump.index.date)[0].days
            # 筛选出所有向上跳空的
            up_jumps = jumps[(jumps.jump == 1)]
            if up_jumps.shape[0] > 0:
                # 筛选最后一次向上跳空的情况
                last_up_jump = up_jumps.iloc[-1:]
                # 跳空能力* last_up_jump.jump.values[0] 转换成有方向的
                jump_up_power = last_up_jump.jump_power.values[0] * last_up_jump.jump.values[0]
                # 向上跳空距离买入或卖出的时间间隔
                diff_up_days = (kl_pd.iloc[day_ind: day_ind + 1].index.date - last_up_jump.index.date)[0].days

        # 标准化跳空特征特征值
        jump_down_power = round(jump_down_power, 3)
        jump_up_power = round(jump_up_power, 3)

        # 跳空特征键值对字典添加跳空特征key和对应的跳空特征值
        jump_dict['{}jump_down_power'.format(key_prefix)] = jump_down_power
        jump_dict['{}diff_down_days'.format(key_prefix)] = diff_down_days

        jump_dict['{}jump_up_power'.format(key_prefix)] = jump_up_power
        jump_dict['{}diff_up_days'.format(key_prefix)] = diff_up_days

        return jump_dict


class AbuFeatureSnapshot(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """
        快照特征，支持买入，卖出 abupy.env.g_enable_take_kl_snapshot开关控制特征是否生成，
        生成的走势图在～/abu/data/save_png/今天的日期/目录下
    """

    def __init__(self):
        """
            默认60日价格做为快照，如外部修改，直接使用类似如下：
                abupy.feature.g_take_snap_shot_xd = 30
        """
        self.take_snap_shot_xd = g_take_snap_shot_xd
        self.snap_key = 'snap'

    def get_feature_keys(self, buy_feature):
        """
        返回对应的快照特征key值
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 快照特征的键值对字典中的key序列
        """
        return ['{}{}'.format(self.feature_prefix(buy_feature=buy_feature), self.snap_key)]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造快照特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造快照特征的键值对字典
        """

        # 快照生成时间字符串
        tt = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
        # 生成64位随机字符串
        rn = ABuStrUtil.create_random_with_alpha(64)
        # 快照文件名确定，由于会使用多任务并行，所以加入进程id，和64位随机数，避免产生文件名冲突
        snap_fn = '{}_{}_{}'.format(tt, os.getpid(), rn)

        if day_ind - self.take_snap_shot_xd >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            snap_window_pd = kl_pd[day_ind - self.take_snap_shot_xd + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-xd:].close，否则取combine_kl_pd所有交易收盘价格
            snap_window_pd = combine_kl_pd[-self.take_snap_shot_xd:] \
                if combine_kl_pd.shape[0] > self.take_snap_shot_xd else combine_kl_pd

        # 模块设置，绘制k线图只绘制价格曲线，不绘制成交量
        ABuMarketDrawing.g_only_draw_price = True
        # 通过特征周期时间序列snap_window_pd做为参数，绘制交易k线图快照，保存在本地，详ABuMarketDrawing.plot_candle_form_klpd
        ABuMarketDrawing.plot_candle_form_klpd(snap_window_pd, save=True, name=snap_fn)
        # 快照特征键值对字典添加快照特征key和对应的快照特征值
        snap_dict = {'{}{}'.format(self.feature_prefix(buy_feature=buy_feature),
                                   self.snap_key): ABuMarketDrawing.save_dir_name() + snap_fn + '.png'}
        return snap_dict


"""用户可扩展自定义特征"""
_g_extend_feature_list = list()


def append_user_feature(feature, check=True):
    """
    外部设置扩展feature接口
    :param feature: 可以是feature class类型，也可以是实例化后的feature object
    :param check: 是否检测feature是AbuFeatureBase实例
    :return:
    """

    if isinstance(feature, six.class_types):
        # 暂时认为所有feature的实例化不需要参数，如需要也可添加＊args
        feature_obj = feature()
    else:
        # 如果不是类直接赋值
        feature_obj = feature

    # check检测feature_obj是不是AbuFeatureBase的子类实例对象
    if check and not isinstance(feature_obj, AbuFeatureBase):
        raise TypeError('feature must a isinstance AbuFeatureBase!!!')

    # 添加到用户可扩展自定义特征序列中
    _g_extend_feature_list.append(feature_obj)


def clear_user_feature():
    """将用户可扩展自定义特征序列清空"""
    global _g_extend_feature_list
    _g_extend_feature_list = list()


class AbuMlFeature(object):
    """特征对外统一接口类，负责管理构建内部特征，用户扩展特征，提供买入卖出因子的交易特征生成，转换接口"""

    def __init__(self):
        """实例化 内置特征对象＋用户扩展自定义特征对象"""

        # 内置特征实例化
        self.features = [AbuFeatureDeg(), AbuFeaturePrice(), AbuFeatureWave(), AbuFeatureAtr(), AbuFeatureJump()]
        if ABuEnv.g_enable_take_kl_snapshot:
            # 快照特征比较特殊，默认不开启，因为大规模回测比较耗时等开销，故开关控制
            self.features.append(AbuFeatureSnapshot())

        # 用户扩展自定义特征对象extend到特征对象序列self.features中
        if len(_g_extend_feature_list) > 0:
            # 这里不再次check了，因为append_user_feature时已经做过，直接extend
            self.features.extend(_g_extend_feature_list)

    def make_feature_dict(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        提供买入卖出因子构建交易特征的接口，使用见AbuFactorBuyBase.make_buy_order_ml_feature
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        """

        if not ABuEnv.g_enable_ml_feature:
            # 需要env中开启特征生成，否则不生成交易特征
            return None
        ml_feature_dict = {}
        # 根据buy_feature，过滤出特征序列中支持的特征序列子集
        features = list(
            filter(lambda f: f.support_buy_feature() if buy_feature else f.support_sell_feature(), self.features))
        for feature in features:
            # 迭代特征序列对象，特征对象统一使用calc_feature接口生成自己的特征，结果特征update到总特征字典ml_feature_dict中
            ml_feature_dict.update(feature.calc_feature(kl_pd, combine_kl_pd, day_ind, buy_feature))
        return ml_feature_dict

    def _get_unzip_feature_keys(self, buy_feature):
        """
        根据buy_feature，过滤出特征序列中支持的特征序列子集，迭代特征序列对象，
        使用特征对象的get_feature_keys方法获取特征的key序列
        :param buy_feature: 是否是买入特征构造（bool）
        """
        feature_key_list = []
        # 根据buy_feature，过滤出特征序列中支持的特征序列子集
        features = list(
            filter(lambda f: f.support_buy_feature() if buy_feature else f.support_sell_feature(), self.features))

        # 迭代特征序列对象，使用特征对象的get_feature_keys方法获取特征的key序列
        for feature in features:
            feature_key_list.extend(feature.get_feature_keys(buy_feature))
        return feature_key_list

    def unzip_ml_feature(self, orders_pd):
        """
        ABuTradeExecute中make_orders_pd使用，将order中dict字典形式的特征解压拆解为独立的
        pd.DataFrame列，即一个特征key，对应一个列
        :param orders_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :return:
        """
        if ABuEnv.g_enable_ml_feature:
            features_keys = list()
            # 收集买入特征keys
            features_keys.extend(self._get_unzip_feature_keys(True))
            # 收集卖出特征keys
            features_keys.extend(self._get_unzip_feature_keys(False))

            # from ..UtilBu.ABuDTUtil import except_debug
            # @except_debug
            def map_str_dict(order, key):
                if order.sell_type == 'keep' and key.startswith('sell_'):
                    # 针对卖出特征值，如果单子keep状态，即没有特征值
                    return np.nan

                if not isinstance(order.ml_features, dict):
                    # 低版本pandas dict对象取出来会成为str
                    map_ast = ast.literal_eval(order.ml_features)[key]
                else:
                    map_ast = order.ml_features[key]

                return map_ast

            for fk in features_keys:
                # 迭代所有key，fk做为pd.DataFrame对象orders_pd的新列名，
                orders_pd[fk] = orders_pd.apply(map_str_dict, axis=1, args=(fk,))


class AbuFeatureDegExtend(AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """示例添加新的视角来录制比赛，角度特征，支持买入，卖出"""

    def __init__(self):
        """20, 40, 60, 90, 120日走势角度特征"""
        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.deg_keys = frozenset([10, 30, 50, 90, 120])

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有走势角度特征feature的列名称定, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 角度特征的键值对字典中的key序列
        """
        return ['{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature), dk) for dk in self.deg_keys]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造拟合角度特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造角度特征的键值对字典
        """
        # 返回的角度特征键值对字典
        deg_dict = {}
        for dk in self.deg_keys:
            # 迭代预设角度周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                deg_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                deg_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[0] > dk else combine_kl_pd.close

            # 使用截取特征交易周期收盘价格deg_close做为参数，通过calc_regress_deg计算趋势拟合角度
            ang = ABuRegUtil.calc_regress_deg(deg_close, show=False)
            # 标准化拟合角度值
            ang = 0 if np.isnan(ang) else round(ang, 3)
            # 角度特征键值对字典添加拟合角度周期key和对应的拟合角度值
            deg_dict['{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature), dk)] = ang
        return deg_dict



# -*- encoding:utf-8 -*-
"""卖出因子参数以及选择图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import abstractmethod

import ipywidgets as widgets

from ..WidgetBu.ABuWGBase import WidgetFactorBase, WidgetFactorManagerBase
from ..WidgetBu.ABuWGBFBase import BFSubscriberMixin

__author__ = '阿布'
__weixin__ = 'abu_quant'


class SellFactorWGManager(WidgetFactorManagerBase):
    """卖出策略组织类"""

    def _init_widget(self):
        """构建内置的卖出策略可视化组件，构造出self.factor_box"""

        from ..WidgetBu.ABuWGSellFactor import SellXDWidget, SellAtrNStopWidget, SellCloseAtrNWidget
        from ..WidgetBu.ABuWGSellFactor import SellPreAtrNWidget, SellDMWidget, SellNDWidget

        self.sf_array = []
        self.sf_array.append(SellAtrNStopWidget(self))
        self.sf_array.append(SellCloseAtrNWidget(self))
        self.sf_array.append(SellPreAtrNWidget(self))
        self.sf_array.append(SellXDWidget(self))
        self.sf_array.append(SellDMWidget(self))
        self.sf_array.append(SellNDWidget(self))

        # sf() call用widget组list
        children = [sf() for sf in self.sf_array]
        if self.scroll_factor_box:
            self.factor_box = widgets.Box(children=children,
                                          layout=self.factor_layout)
        else:
            # 一行显示两个，2个为一组，组装sub_children_group序列,
            sub_children_group = self._sub_children(children, len(children) / 2)
            sub_children_box = [widgets.HBox(sub_children) for sub_children in sub_children_group]
            self.factor_box = widgets.VBox(sub_children_box)

        # 买入因子是特殊的存在，都需要买入因子的全局数据
        self.buy_factor_manger = None

    def register(self, buy_factor_manger):
        """卖出manager内部因子注册接收买入因子添加的改变"""
        self.buy_factor_manger = buy_factor_manger
        for sf in self.sf_array:
            self.buy_factor_manger.register_subscriber(sf)


class WidgetFactorSellBase(WidgetFactorBase, BFSubscriberMixin):
    """卖出策略可视化基础类"""

    def __init__(self, wg_manager):
        super(WidgetFactorSellBase, self).__init__(wg_manager)
        self.add = widgets.Button(description=u'添加为全局卖出策略', layout=widgets.Layout(width='98%'),
                                  button_style='info')
        # 添加全局卖出策略指令按钮
        self.add.on_click(self.add_sell_factor)
        # 运行混入的BFSubscriberMixin中ui初始化
        self.subscriber_ui([u'点击\'已添加的买入策略\'框中的买入策略', u'将卖出策略做为买入策略的附属卖出策略'])
        # 买入策略框点击行为：将本卖出策略加到对应的买入策略做为附属
        self.buy_factors.observe(self.add_sell_factor_to_buy_factor, names='value')
        self.accordion.set_title(0, u'添加为指定买入因子的卖出策略')
        self.accordion.selected_index = -1
        self.add_box = widgets.VBox([self.add, self.accordion])

        self._init_widget()

    @abstractmethod
    def _init_widget(self):
        """子类因子界面设置初始化"""
        pass

    @abstractmethod
    def make_sell_factor_unique(self):
        """
            子类因子构建唯一描述以及因子字典
            返回值两个：
            1. 因子构建字典对象
            2. 因子唯一描述
        """
        pass

    # noinspection PyUnusedLocal
    def add_sell_factor(self, bt):
        """对应按钮添加全局策略"""
        # 构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key
        factor_dict, factor_desc_key = self.make_sell_factor_unique()
        self.wg_manager.add_factor(factor_dict, factor_desc_key)

    def add_sell_factor_to_buy_factor(self, select):
        """对应按钮添加策略到指定买入策略中"""
        self.add_to_buy_factor(select, self.make_sell_factor_unique, 'sell_factors')




# -*- encoding:utf-8 -*-
"""股票基本信息图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import ipywidgets as widgets

from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EDataCacheType, EMarketDataFetchMode, EMarketSourceType
from ..WidgetBu.ABuWGBase import WidgetBase
from ..UtilBu import ABuFileUtil

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetEnvSetMixin(object):
    """
        使用混入而不要做为上层widget拥有的模式，可为多个上层使用
        便于上层widgte使用self去获取设置，统一上层使用
        混入类：基础env设置：
        1. 沙盒模式与实时
        2. csv模式与hdf5模式
        3. 数据获取模式
        4. 数据源切换
    """

    # noinspection PyProtectedMember
    def init_env_set_ui(self):
        """构建基础env widget ui return widgets.VBox"""
        """沙盒数据与开放数据模式切换"""
        self.date_mode = widgets.RadioButtons(
            options=[u'沙盒数据模式', u'开放数据模式'],
            value=u'沙盒数据模式' if ABuEnv._g_enable_example_env_ipython else u'开放数据模式',
            description=u'数据模式:',
            disabled=False
        )
        self.date_mode.observe(self.on_data_mode_change, names='value')

        set_mode_label_tip1 = widgets.Label(u'下面的设置|缓存模式|联网模式|数据源：',
                                            layout=widgets.Layout(width='300px', align_items='stretch'))
        set_mode_label_tip2 = widgets.Label(u'只在开放数据模式下生效',
                                            layout=widgets.Layout(width='300px', align_items='stretch'))

        """csv模式与hdf5模式模式切换"""
        self.store_mode_dict = {EDataCacheType.E_DATA_CACHE_CSV.value: u'csv模式(推荐)',
                                EDataCacheType.E_DATA_CACHE_HDF5.value: u'hdf5模式'}
        self.store_mode = widgets.RadioButtons(
            options=[u'csv模式(推荐)', u'hdf5模式'],
            value=self.store_mode_dict[ABuEnv.g_data_cache_type.value],
            description=u'缓存模式:',
            disabled=False
        )
        self.store_mode.observe(self.on_data_store_change, names='value')

        """数据获取模式模式切换"""
        self.fetch_mode_dict = {EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL.value: u'本地数据模式(推荐)',
                                EMarketDataFetchMode.E_DATA_FETCH_NORMAL.value: u'本地网络结合模式',
                                EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET.value: u'强制全部使用网络'}
        self.fetch_mode = widgets.RadioButtons(
            options=[u'本地数据模式(推荐)', u'本地网络结合模式', u'强制全部使用网络'],
            value=self.fetch_mode_dict[ABuEnv.g_data_fetch_mode.value],
            description=u'联网模式:',
            disabled=False
        )
        self.fetch_mode.observe(self.on_fetch_mode_change, names='value')

        """数据源进行切换"""
        self.data_source_accordion = widgets.Accordion()
        self.date_source_dict = {EMarketSourceType.E_MARKET_SOURCE_bd.value: u'百度数据源(美股，A股，港股)',
                                 EMarketSourceType.E_MARKET_SOURCE_tx.value: u'腾讯数据源(美股，A股，港股)',
                                 EMarketSourceType.E_MARKET_SOURCE_nt.value: u'网易数据源(美股，A股，港股)',
                                 EMarketSourceType.E_MARKET_SOURCE_sn_us.value: u'新浪美股(美股)',
                                 EMarketSourceType.E_MARKET_SOURCE_sn_futures.value: u'新浪国内期货(国内期货)',
                                 EMarketSourceType.E_MARKET_SOURCE_sn_futures_gb.value: u'新浪国际期货(国际期货)',
                                 EMarketSourceType.E_MARKET_SOURCE_hb_tc.value: u'比特币，莱特币'}
        self.date_source = widgets.RadioButtons(
            options=[u'百度数据源(美股，A股，港股)', u'腾讯数据源(美股，A股，港股)', u'网易数据源(美股，A股，港股)',
                     u'新浪美股(美股)', u'新浪国内期货(国内期货)', u'新浪国际期货(国际期货)',
                     u'比特币，莱特币'],
            value=self.date_source_dict[ABuEnv.g_market_source.value],
            description=u'数据源:',
            disabled=False
        )
        self.date_source.observe(self.on_date_source_change, names='value')
        source_label_tip1 = widgets.Label(u'内置的数据源仅供学习使用',
                                          layout=widgets.Layout(width='300px', align_items='stretch'))
        source_label_tip2 = widgets.Label(u'abupy提供了接入外部数据源的接口和规范',
                                          layout=widgets.Layout(width='300px', align_items='stretch'))
        source_label_tip3 = widgets.Label(u'详阅读github上教程第19节的示例',
                                          layout=widgets.Layout(width='300px', align_items='stretch'))

        other_data_set_box = widgets.VBox([self.fetch_mode, source_label_tip1, self.date_source, source_label_tip2,
                                           source_label_tip3, self.store_mode])

        self.data_source_accordion.children = [other_data_set_box]
        self.data_source_accordion.set_title(0, u'缓存模式|联网模式|数据源')
        self.data_source_accordion.selected_index = -1

        mdm_box = widgets.VBox([self.date_mode, set_mode_label_tip1, set_mode_label_tip2, self.data_source_accordion])

        if ABuEnv._g_enable_example_env_ipython:
            # 非沙盒数据下数据存贮以及数据获取模式切换才生效
            self.store_mode.disabled = True
            self.fetch_mode.disabled = True

        return mdm_box

    def on_data_mode_change(self, change):
        """沙盒与非沙盒数据界面操作转换"""
        if change['new'] == u'沙盒数据模式':
            ABuEnv.enable_example_env_ipython(show_log=False)
            self.store_mode.disabled = True
            self.fetch_mode.disabled = True
            self.data_source_accordion.selected_index = -1
        else:
            if ABuFileUtil.file_exist(ABuEnv.g_project_kl_df_data_csv) and \
                            len(os.listdir(ABuEnv.g_project_kl_df_data_csv)) > 5000:
                # 如果有很多缓存数据，从沙盒改变依然网络模式是本地模式
                ABuEnv._g_enable_example_env_ipython = False
            else:
                # 如果没有很多缓存从沙盒改到开放一起改变网络模式本地网络结合
                ABuEnv.disable_example_env_ipython(show_log=False)

            self.store_mode.disabled = False
            self.fetch_mode.disabled = False
            self.data_source_accordion.selected_index = 0
        # 需要数据获取模式界面进行同步更新
        self.fetch_mode.value = self.fetch_mode_dict[ABuEnv.g_data_fetch_mode.value]

    def on_data_store_change(self, change):
        """数据存储模式界面操作进行改变"""
        store_mode_dict = {self.store_mode_dict[sk]: EDataCacheType(sk) for sk in self.store_mode_dict}
        # 改变设置值
        ABuEnv.g_data_cache_type = store_mode_dict[change['new']]

    def on_fetch_mode_change(self, change):
        """数据获取模式界面操作进行改变"""
        fetch_mode_dict = {self.fetch_mode_dict[fm]: EMarketDataFetchMode(fm) for fm in self.fetch_mode_dict}
        # 改变设置值
        ABuEnv.g_data_fetch_mode = fetch_mode_dict[change['new']]

    def on_date_source_change(self, change):
        """数据源界面操作进行改变"""
        date_source_dict = {self.date_source_dict[ds]: EMarketSourceType(ds) for ds in self.date_source_dict}
        # 改变设置值
        ABuEnv.g_market_source = date_source_dict[change['new']]


class WidgetTimeModeMixin(object):
    """
        使用混入而不要做为上层widget拥有的模式，可为多个上层使用
        便于上层widgte使用self去获取设置，统一上层使用
        混入类：基础时间模式设置：
        1. 年数模式
        2. 开始结束模式
    """

    # noinspection PyProtectedMember
    def init_time_mode_ui(self):
        """构建基础env widget ui return widgets.VBox"""
        # 回测时间模式
        self.time_mode = widgets.RadioButtons(
            options={u'使用{}年数'.format(self.time_mode_str()): 0,
                     u'使用{}开始结束日期'.format(self.time_mode_str()): 1},
            value=0,
            description=u'时间模式:',
            disabled=False
        )
        self.time_mode.observe(self.on_time_mode_change, names='value')

        # 年数模式
        self.run_years = widgets.BoundedIntText(
            value=2,
            min=1,
            max=6,
            step=1,
            description=u'{}年数:'.format(self.time_mode_str()),
            disabled=False
        )
        # 开始结束模式
        self.start = widgets.Text(
            value='2014-07-26',
            placeholder=u'年-月-日',
            description=u'开始日期:',
            disabled=False
        )
        self.end = widgets.Text(
            value='2016-07-26',
            placeholder=u'年-月-日',
            description=u'结束日期:',
            disabled=False
        )
        self.run_years.disabled = False
        self.start.disabled = True
        self.end.disabled = True

        return widgets.VBox([self.time_mode, self.run_years, self.start, self.end])

    def on_time_mode_change(self, change):
        """切换使用年数还是起始，结束时间做为回测参数"""
        if change['new'] == 0:
            self.run_years.disabled = False
            self.start.disabled = True
            self.end.disabled = True
        else:
            self.run_years.disabled = True
            self.start.disabled = False
            self.end.disabled = False

    def time_mode_str(self):
        """子类实现返回一个字符串代表时间设置的意义eg：回测，分析"""
        raise NotImplementedError('NotImplementedError time_mode_str!')


class WidgetRunTT(WidgetBase, WidgetEnvSetMixin, WidgetTimeModeMixin):
    """基础设置界面：初始资金，回测开始，结束周期，参考大盘等"""

    def __init__(self):
        """初始化基础回测设置界面"""
        # 初始资金
        self.cash = widgets.BoundedIntText(
            value=1000000,
            min=10000,
            max=999999999,
            step=1,
            description=u'初始资金:',
            disabled=False
        )
        tm_box = self.init_time_mode_ui()
        mdm_box = self.init_env_set_ui()
        self.widget = widgets.VBox([self.cash, tm_box, mdm_box])

    def time_mode_str(self):
        """实现混入WidgetTimeModeMixin，声明时间模块代表回测"""
        return u'回测'
